/***  Module: DISPUTIL
      Device Independent General Purpose Graphics Routines

    Includes line drawing and cursor routines.
***/

#define __MSC

/* 10/6 Changed cursor routines for cursor control without numlock - mdm*/

/* * * * INCLUDE files * * * */

#include  <conio.h>
#include  <stdio.h>
#include  <string.h>
#include  <stdlib.h>
#include  "imdef.h"
#include  "imdisp.h"
#include  "dispio.h"
#include  "refresh.h"
#include  "textutil.h"
#include  "palutil.h"

/* * * * External functions * * * */
extern void pascal far mouses( int *, int *, int *, int * );
extern int use_mouse;


/* * * * Function declarations * * * */

int  DrawLine ( int, int, int, int, int);
int  DrawBox (int, int, int, int, int );
int  RemoveCursor (void);
int  PlaceCursor (int, int, int);
int  MoveCursorInput (int *, int *, unsigned char *);
int  MoveCursor (int *, int *);
int  InitDisplay (void);
void FrameBox (int, int, int, int, int, int);
void RubberBand (int, int, int *, int *);

/* * * * Global Variables * * * */

unsigned char  CursorShape[CURSORSIZE][CURSORSIZE]
                            = { {0,0,0,0,1,0,0,0,0},
                                {0,0,0,0,1,0,0,0,0},
                                {0,0,0,0,1,0,0,0,0},
                                {0,0,0,0,1,0,0,0,0},
                                {1,1,1,1,1,1,1,1,1},
                                {0,0,0,0,1,0,0,0,0},
                                {0,0,0,0,1,0,0,0,0},
                                {0,0,0,0,1,0,0,0,0},
                                {0,0,0,0,1,0,0,0,0} };

int            CursorOn;                   /* cursor flag */
int            CursorLine, CursorSample;   /* cursor location */
int            CursorInc;                  /* size of cursor steps */
unsigned char  CursorPatch[CURSORSIZE][CURSORSIZE];
                 /* contains display pixels displaced by cursor */


int DrawLine (int x1, int y1, int x2, int y2, int color)

/***  DrawLine draws a line of pixels from the first point to the
    second point with the desired DN value.  Clipping is not performed.

       Parameter  Type   Description
         x1       int    the line coordinate of the first point
         y1       int    the sample coordinate of the first point
         x2       int    the line coordinate of the second point
         y2       int    the sample coordinate of the second point
         color    int    the DN value of the drawn pixels
***/

{
    int  x, y, delx, dely, absdelx, absdely, rem, address;
    unsigned char buffer[1];
    buffer[0]= color;
    delx = x2 - x1;
    dely = y2 - y1;
    absdelx = abs(delx);
    absdely = abs(dely);
    rem = 0;

    if (absdely < absdelx)
    {
       y = y1;
       if (delx > 0)
       {
           if (dely > 0)
           {
               for (x = x1; x <= x2; x++)
               {
                   rem += absdely;
                   if (rem > absdelx)
                   {
                       rem -= absdelx;
                       y++;
                   }
                   if ((IntoMem) && (RefreshLines >= x))
                       PutRefresh(buffer,x,y,1);
                   else
                       WritePixel (x,y, color);
               }
           }
           else
           {
               for (x = x1;  x <= x2;  x++)
               {
                   rem += absdely;
                   if (rem > absdelx)
                   {
                       rem -= absdelx;
                       y--;
                   }
                   if ((IntoMem) && (RefreshLines >= x))
                       PutRefresh(buffer,x,y,1);
                   else
                       WritePixel (x,y, color);
               }
           }
       }

       else
       {
           if (dely > 0)
           {
               for (x = x1;  x >= x2;  x--)
               {
                   rem += absdely;
                   if (rem > absdelx)
                   {
                       rem -= absdelx;
                       y++;
                   }
                   if ((IntoMem) && (RefreshLines >= x))
                       PutRefresh(buffer,x,y,1);
                   else
                       WritePixel (x,y, color);
               }
           }
           else
           {
               for (x = x1;  x >= x2;  x--)
               {
                   rem += absdely;
                   if (rem > absdelx)
                   {
                       rem -= absdelx;
                       y--;
                   }
                   if ((IntoMem) && (RefreshLines >= x))
                       PutRefresh(buffer,x,y,1);
                   else
                       WritePixel (x,y, color);
               }
           }
       }
    }

    else
    {
       x = x1;
       if (dely > 0)
       {
           if (delx > 0)
           {
               for (y = y1;  y <= y2;  y++)
               {
                   rem += absdelx;
                   if (rem > absdely)
                   {
                       rem -= absdely;
                       x++;
                   }
                   if ((IntoMem) && (RefreshLines >= x))
                       PutRefresh(buffer,x,y,1);
                   else
                       WritePixel (x,y, color);
               }
           }
           else
           {
               for (y = y1;  y <= y2;  y++)
               {
                   rem += absdelx;
                   if (rem > absdely)
                   {
                       rem -= absdely;
                       x--;
                   }
                   if ((IntoMem) && (RefreshLines >= x))
                       PutRefresh(buffer,x,y,1);
                   else
                       WritePixel (x,y, color);
               }
           }
       }

       else
       {
           if (delx > 0)
           {
               for (y = y1;  y >= y2;  y--)
               {
                   rem += absdelx;
                   if (rem > absdely)
                   {
                       rem -= absdely;
                       x++;
                   }
                   if ((IntoMem) && (RefreshLines >= x))
                       PutRefresh(buffer,x,y,1);
                   else
                       WritePixel (x,y, color);
               }
           }
           else
           {
               for (y = y1;  y >= y2; y--)
               {
                   rem += absdelx;
                   if (rem > absdely)
                   {
                       rem -= absdely;
                       x--;
                   }
                   if ((IntoMem) && (RefreshLines >= x))
                       PutRefresh(buffer,x,y,1);
                   else
                       WritePixel (x,y, color);
               }
           }
       }
    }
}

int DrawBox (int line, int samp, int lsize, int ssize, int color)

/*** DrawBox draws a solid rectangle on the screen.
   Used for erasing parts of the screen.

    Parameter   Type     Description
     line       int      starting line coordinate of rectangle
     samp       int      starting sample coordinate of rectangle
     lsize      int      number of lines in rectangle
     ssize      int      number of samples in rectangle
     color      int      the DN value to draw the box
***/

{
    int    l, s;
    unsigned char  linebuf[MAXDISPNS];

    for (s = 0;  s < ssize;  s++)
        linebuf[s] = color;
    for (l = line;  l < line+lsize;  l++)
    {
       if ((IntoMem) && (RefreshLines >= l))
           PutRefresh(linebuf,l,samp,ssize);

       DisplayLine (linebuf, l, samp, ssize);
     }
}




int RemoveCursor(void)

/***  RemoveCursor removes the cursor, if it is on, from the display
    screen, replacing the original pixel values.
***/
{
    int    lcorn, scorn, l, s, DN;
    int    ls, le, ss, se;

    if (CursorOn)
    {
        lcorn = CursorLine - (CURSORSIZE / 2) - 1;
        scorn = CursorSample - (CURSORSIZE / 2) - 1;
        ls = (lcorn < 0) ?  1 - lcorn : 1;
        le = (lcorn > dispnl-CURSORSIZE) ?  dispnl-lcorn : CURSORSIZE;
        ss = (scorn < 0) ?  1 - scorn : 1;
        se = (scorn > dispns-CURSORSIZE) ?  dispns-scorn : CURSORSIZE;

        for (l = ls;  l <= le;  l++)
            for (s = ss;  s <= se;  s++)
                if (CursorShape[l-1][s-1] > 0)
                {
                    DN = CursorPatch[l-1][s-1];
                    WritePixel (lcorn+l, scorn+s, DN);
                }
        CursorOn = 0;
    }
}


int PlaceCursor (int line, int sample, int cursDN)

/***  PlaceCursor puts the cursor, if it is not already on, on the
    screen, while storing the original pixel values.
     Parameter     Type     Description
      line         int      the line coordinate to put the cursor at
      sample       int      the sample coordinate to put the cursor at
      cursDN       int      the DN value of the cursor

***/

{
    int   lcorn, scorn, l, s, DN;
    int   ls, le, ss, se, ld, sd;

    if (!CursorOn)
    {
        lcorn = line - (CURSORSIZE / 2) - 1;
        scorn = sample - (CURSORSIZE / 2) - 1;
        ls = (lcorn < 0) ?  1 - lcorn : 1;
        le = (lcorn > dispnl-CURSORSIZE) ?  dispnl-lcorn : CURSORSIZE;
        ss = (scorn < 0) ?  1 - scorn : 1;
        se = (scorn > dispns-CURSORSIZE) ?  dispns-scorn : CURSORSIZE;

        for (l = ls;  l <= le;  l++)
            for (s = ss;  s <= se;  s++)
                if (CursorShape[l-1][s-1] > 0)
                {
                    ld = lcorn + l;
                    sd = scorn + s;
                    ReadPixel (ld, sd, &DN);
                    CursorPatch[l-1][s-1] = DN;
                    WritePixel (ld, sd, cursDN );
                }
        CursorLine = line;
        CursorSample = sample;
        CursorOn = 1;
    }
}

int MoveCursorInput (int *p_line, int *p_sample, unsigned char *ch)

/***  MoveCursorInput puts the cursor at the given place, waits
    for a character from the keyboard, performs the cursor command
    for the input character, and removes the cursor.
     Parameter     Type     Description
      p_line     int ptr     the cursor line coordinate
      p_sample   int ptr     the cursor sample coordinate
      ch         char ptr    the character input
***/

{
    int    DN, center, len, m1, m2, m3, m4;
    char   dispstr[32];

    PlaceCursor (*p_line, *p_sample, numDN-1);

    /* moved display of curs here - otherwise 1 behind mdm 3/88 */
    center = CURSORSIZE/2;
    DN     = CursorPatch[center][center];

    sprintf (dispstr, "%4d %4d %3d", *p_line, *p_sample, DN);
    StatusLine(2,dispstr);

/* Mouse functions added by AW3, 5/91 from modified v4.5
   Original comment follows                              */
/* Changed 9/26/90 to include mouse support - mwm */
    if ( use_mouse )
    {
    /* Mouse? */
       *ch = 0x00;                              /* Clear char */
       m1 = 5;                                  /* Button function */
       m2 = 0;                                  /* Left button */
       mouses( &m1, &m2, &m3, &m4 );            /* Clear left counts */
       do
       {                                        /* Loop */
           m1 = 11;                             /* Motion function */
           mouses( &m1, &m2, &m3, &m4 );        /* Read motion */
           if ( m3 > 0 )                        /* Right? */
               *p_sample = min(*p_sample+m3, dispns);
           else if ( m3 < 0 )                   /* Left? */
               *p_sample = max(*p_sample+m3, 1);
           if ( m4 < 0 )                        /* Up? */
               *p_line = max(*p_line+m4, 1);
           else if ( m4 > 0 )                   /* Down? */
               *p_line = min(*p_line+m4, dispnl);
           if ( (m3 != 0) || (m4 != 0) )        /* Motion? */
              break;                            /* Yes - finished */
           m1 = 5;                              /* Button function */
           m2 = 0;                              /* Left button */
           mouses( &m1, &m2, &m3, &m4 );        /* Check button */
           if ( m2 > 0 )
           {                                    /* Pressed? */
               *ch = '.';                       /* Yes - set end char */
               break;                           /* and quit */
           }
       } while ( !kbhit() );                    /* Until key pressed */
       if ( !kbhit() )
       {                                        /* Key pressed? */
           RemoveCursor();                      /* No - finished */
           return(DN);
       }
    }

  if ((*ch = getch()) == 0)  /* mdm 10/6 */
      *ch = 0x80 | getch();

    switch (*ch)
    {
    case UP_ARROW:
    case '8' :
             *p_line = max(*p_line-CursorInc, 1);
             break;
    case DOWN_ARROW:
    case '2' :
             *p_line = min(*p_line+CursorInc, dispnl);
             break;
    case LEFT_ARROW:
    case '4' :
             *p_sample = max(*p_sample-CursorInc, 1);
             break;
    case RIGHT_ARROW:
    case '6' :
             *p_sample = min(*p_sample+CursorInc, dispns);
             break;
    case UP_LEFT_ARROW:
    case '7' :
             *p_line = max(*p_line-CursorInc, 1);
             *p_sample = max(*p_sample-CursorInc, 1);
             break;
    case UP_RIGHT_ARROW:
    case '9' :
             *p_line = max(*p_line-CursorInc, 1);
             *p_sample = min(*p_sample+CursorInc, dispns);
             break;
    case DOWN_RIGHT_ARROW:
    case '3' :
             *p_line = min(*p_line+CursorInc, dispnl);
             *p_sample = min(*p_sample+CursorInc, dispns);
             break;
    case DOWN_LEFT_ARROW:
    case '1' :
             *p_line = min(*p_line+CursorInc, dispnl);
             *p_sample = max(*p_sample-CursorInc, 1);
             break;
    case '-' :
    case '_' : /* accept as unshifted minus */
             CursorInc = max(CursorInc / 2, 1);
             break;
    case '=' : /* accept as unshifted plus */ 
    case '+' :
             CursorInc = min(2*CursorInc, dispnl / 4);
             break;
    }

    RemoveCursor();
    return(DN);
}

int MoveCursor (int * p_line, int * p_sample)

/***  MoveCursor performs the cursor mode operation.
    The user moves the cursor around using the numeric keypad.
    The cursor moves in discrete steps whose size can be changed.
    The final cursor position is returned.

    The following cursor command are performed:

           Character      Action
              8           cursor up one step
              9           cursor up and right one step
              6           cursor right one step
              3           cursor down and right one step
              2           cursor down one step
              1           cursor down and left one step
              4           cursor left one step
              7           cursor up and left one step
              5           cursor position and correspoding pixel value
                              display at upper right of screen
              -           cursor step size decreased by factor of 2
              +           cursor step size increased by factor of 2
              .           exits cursor mode

     Parameter     Type      Description
      p_line     int ptr     the cursor line coordinate
      p_sample   int ptr     the cursor sample coordinate
***/

{
    unsigned  char   ch;  /* mdm 10/6 */
    int    DN;

    *p_line = CursorLine;
    *p_sample = CursorSample;
    do
    {
       DN = MoveCursorInput (p_line, p_sample, &ch);
       if (ch == RETURN || ch == ESCAPE)
           ch = '.';
    }
    while (ch != '.');
    return(DN);
}

int InitDisplay(void)

/*** InitDisplay turns the display device on, loads the default palette,
    initializes the global variables, and loads the font arrays.
***/

{
    int    i;
    char   *devname;

    DisplayDevice = 0;

    if ((devname = getenv("IMDISP")) != NULL)
      {
       if (stricmp(devname, "cga"       ) == 0) DisplayDevice = CGA200;
       if (stricmp(devname, "ega480"    ) == 0) DisplayDevice = EGA480;
       if (stricmp(devname, "ega"       ) == 0) DisplayDevice = EGA350;
       if (stricmp(devname, "pga"       ) == 0) DisplayDevice = PGA;
       if (stricmp(devname, "vga320"    ) == 0) DisplayDevice = VGA200;
       if (stricmp(devname, "vga"       ) == 0) DisplayDevice = VGA480;
       if (stricmp(devname, "orchid1024") == 0) DisplayDevice = ORCHID768;
       if (stricmp(devname, "orchid800" ) == 0) DisplayDevice = ORCHID600;
       if (stricmp(devname, "orchid"    ) == 0) DisplayDevice = ORCHID480;
       if (stricmp(devname, "evga640"   ) == 0) DisplayDevice = EVGA640;
       if (stricmp(devname, "evga512"   ) == 0) DisplayDevice = EVGA512;
       if (stricmp(devname, "evga800"   ) == 0) DisplayDevice = EVGA800;
       if (stricmp(devname, "ati640"    ) == 0) DisplayDevice = ATI640;
       if (stricmp(devname, "ati800"    ) == 0) DisplayDevice = ATI800;
       if (stricmp(devname, "ati1024"   ) == 0) DisplayDevice = ATI1024;
       if (stricmp(devname, "paradise"  ) == 0) DisplayDevice = PARADISE;
       if (stricmp(devname, "trident"   ) == 0) DisplayDevice = TRIDENT;
       if (stricmp(devname, "vesa"      ) == 0) DisplayDevice = VESA;
       if (stricmp(devname, "vesa800"   ) == 0) DisplayDevice = VESA800;
       if (stricmp(devname, "vesa640"   ) == 0) DisplayDevice = VESA640;
       if (stricmp(devname, "ev629"     ) == 0) DisplayDevice = EV629;
       if (stricmp(devname, "test"      ) == 0) DisplayDevice = TEST;
#ifdef __TURBOC__
       if (stricmp(devname, "bios"      ) == 0) DisplayDevice = VGA480;
#else
       if (stricmp(devname, "bios"      ) == 0) DisplayDevice = BIOS;
#endif
      }
    else if (getenv("orchid") != NULL)         DisplayDevice = ORCHID480;
    else if (getenv("ega480") != NULL)         DisplayDevice = EGA480;
    else if (getenv("vga")    != NULL)         DisplayDevice = VGA480;


    DisplayOn();
    WritePalette (DefaultPalette);

    CursorOn   = 0;
    CursorLine = dispnl / 2;    CursorSample = dispns / 2;
    CursorInc  = 16;

    TextLine = TextHeight+5;     TextSample = 1;
}

void FrameBox( int top, int left, int bottom, int right, int color, int flag)
{
       DrawLine( bottom, right, bottom, left, color);
       DrawLine( bottom, left, top, left, color);
       if (flag == TRUE)
       {
           DrawLine( top, left, top, right, color);
           DrawLine( top, right, bottom, right, color);
       }
}

void RubberBand (int line1, int samp1, int * line2, int * samp2)

/***  RubberBand creates a box that is stretched around as the user
    moves the cursor around.
    Written by Ron Baalke - 09/17/91.

    The user moves the cursor around using the numeric keypad.
    The cursor moves in discrete steps whose size can be changed.
    The final cursor position is returned.

    The following cursor command are performed:

           Character      Action
              8           cursor up one step
              9           cursor up and right one step
              6           cursor right one step
              3           cursor down and right one step
              2           cursor down one step
              1           cursor down and left one step
              4           cursor left one step
              7           cursor up and left one step
              5           cursor position and correspoding pixel value
                              display at upper right of screen
              -           cursor step size decreased by factor of 2
              +           cursor step size increased by factor of 2
              .           exits cursor mode

     Parameter     Type      Description
      line1     int         starting line coordinate
      samp1     int         starting sample coordinate
      line2     int ptr     the cursor line coordinate
      samp2     int ptr     the cursor sample coordinate
***/

{
    unsigned  char top_line   [MAXDISPNS];
    unsigned  char bottom_line[MAXDISPNS];
    unsigned  char left_line  [MAXDISPNL];
    unsigned  char right_line [MAXDISPNL];
    int  bottom, left, top, right;
    int  obottom, oleft, otop, oright;
    unsigned  char   ch;
    int box_displayed = FALSE;
    int i,j;

    *line2 = CursorLine;
    *samp2 = CursorSample;
    do
    {
       MoveCursorInput (line2, samp2, &ch);
       if (line1 < *line2)      /* deterimine top, bottom, left, right */
       {                        /* coordinates of the box */
          top = line1;
          bottom = *line2;
       }
       else
       {
          top = *line2;
          bottom = line1;
       }

       if (samp1 < *samp2)
       {
          left = samp1;
          right = *samp2;
       }
       else
       {
          left = *samp2;
          right = samp1;
       }
       if (box_displayed)     /* clear box with lines save previously */
       {
          DisplayLine (top_line   , otop,    oleft, oright-oleft+1);
          DisplayLine (bottom_line, obottom, oleft, oright-oleft+1);
          for (i=otop, j=0; i < otop+obottom; i++,j++)
          {
             WritePixel(i, oleft,  left_line[j]);
             WritePixel(i, oright, right_line[j]);
          }
       }

       /* store away the 4 lines before we draw the box */

       GetLine (top_line, top, left, right-left+1);
       GetLine (bottom_line, bottom, left, right-left+1);
       for (i=top,j=0; i < top+bottom; i++,j++)
       {
          ReadPixel(i, left,  (int *)&left_line[j]);
          ReadPixel(i, right, (int *)&right_line[j]);
       }

       /* draw the box */

       FrameBox(top,left,bottom,right,numDN-1,TRUE);

       /* save coordinates of the box */

       box_displayed = TRUE;
       otop    = top;
       obottom = bottom;
       oleft   = left;
       oright  = right;

       if (ch == RETURN || ch == ESCAPE)
           ch = '.';
    }
    while (ch != '.');

}
