/***  IMDISP module filter.c

    Image Processing Routines

    Contains high level filter routines to perform a number of
    image processing functions such as smoothing and edge
    enhancement.

    Written by Ron Baalke - 05/91
***/

#define __MSC

/* * * * INCLUDE files * * * */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <conio.h>
#include "imdef.h"
#include "imdisp.h"
#include "imdutil.h"
#include "dispio.h"
#include "refresh.h"
#include "display.h"
#include "keywutil.h"
#include "mem.h"
#include "textutil.h"

/* * * * Function declarations * * * */

int DoFilter(void);
int Mean_Filter(void);
int Edge_Filter(void);
int Median_Filter(void);
int Convolution_Filter(void);
int Dark_Filter(int);
int Smear_Filter(void);
int Baalke_Filter(void);
int Unsharp_Filter(void);
int Logarithmic_Filter(void);
int Power_Filter(void);
int Stair_Filter(void);
int HighPass_Filter(void);
int Histogram_Equalization(void);
int DoSlant(void);
int DoMask(void);

/* * * * Global variables * * * */
/*
unsigned char buffer1[MAXDISPNS];
unsigned char buffer2[MAXDISPNS];
unsigned char buffer3[MAXDISPNS];
unsigned char buffer4[MAXDISPNS];
unsigned char buffer5[MAXDISPNS];
unsigned char buffer6[MAXDISPNS];
unsigned char buffer7[MAXDISPNS];
unsigned char outbuffer[MAXDISPNS];
*/
extern unsigned char *buffer1;
extern unsigned char *buffer2;
extern unsigned char *buffer3;
extern unsigned char *buffer4;
extern unsigned char *buffer5;
extern unsigned char *buffer6;
extern unsigned char *buffer7;
extern unsigned char *outbuffer;

/*************************************************************************/
/* DoFilter                                                              */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/* This routine will determine which filter was selected and then call   */
/* the appropriate filter routine.                                       */
/*************************************************************************/

int DoFilter(void)
{
   int flag;
   char dummy[80];

   GetKeywordString (CommandString, "MEA", "",dummy, &flag);
   if (flag != -1)
   {
      Mean_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "EDG", "",dummy, &flag);
   if (flag != -1)
   {
      Edge_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "MED", "",dummy, &flag);
   if (flag != -1)
   {
      Median_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "CON", "",dummy, &flag);
   if (flag != -1)
   {
      Convolution_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "SME", "",dummy, &flag);
   if (flag != -1)
   {
      Smear_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "BAA", "",dummy, &flag);
   if (flag != -1)
   {
      Baalke_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "UNS", "",dummy, &flag);
   if (flag != -1)
   {
      Unsharp_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "LOG", "",dummy, &flag);
   if (flag != -1)
   {
      Logarithmic_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "POW", "",dummy, &flag);
   if (flag != -1)
   {
      Power_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "STA", "",dummy, &flag);
   if (flag != -1)
   {
      Stair_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "HIG", "",dummy, &flag);
   if (flag != -1)
   {
      HighPass_Filter();
      return(0);
   }

   GetKeywordString (CommandString, "HIS", "",dummy, &flag);
   if (flag != -1)
   {
      Histogram_Equalization();
      return(0);
   }

}

/*************************************************************************/
/* Mean_Filter                                                           */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/* This routine will apply a mean filter to the image which has the      */
/* effect of smoothing out the image.  A 3x3 mask is used where the      */
/* middle pixel is replaced by the average of the nine pixels in the     */
/* mask.  Integer math is used to speed up the process.                  */
/*************************************************************************/
int Mean_Filter(void)
{
   int i,j,k;
   int current;
   int new=0;
   unsigned int mean;
   unsigned int sum;
   long offset;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int mask_size;
   int flag;

   /* Apply filter to just the image or the entire screen */

   GetKeywordInteger (CommandString, "MEA", 0, &mask_size, &flag);
   if ((mask_size < 1) || (mask_size > 2)) mask_size = 0;
   if (mask_size == 0)
   {
      mask_size = 3;
      current = 1;
   }
   else if (mask_size == 1)
   {
      mask_size = 5;
      current = 2;
   }
   else
   {
      mask_size = 7;
      current = 3;
   }

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Read in the first three lines of the image */

   GetLine(buffer1,line,sample,length);
   GetLine(buffer2,line+1,sample,length);
   GetLine(buffer3,line+2,sample,length);

   /* If 5x5 mask then read in two more lines */

   if (mask_size == 5)
   {
      GetLine(buffer4,line+3,sample,length);
      GetLine(buffer5,line+4,sample,length);
   }
   else if (mask_size == 7)
   {
      GetLine(buffer6,line+5,sample,length);
      GetLine(buffer7,line+6,sample,length);
   }

   /* Loop through the image line-by-line, but skip the first & last line */

   for (i=line+mask_size/2; i<=line+height- mask_size/2 -1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Store first and last pixel into the outbuffer */

      if (current==0)
      {
         outbuffer[0]        = buffer1[0];
         outbuffer[length-1] = buffer1[length-1];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer1[1];
            outbuffer[length-2] = buffer1[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer1[2];
            outbuffer[length-3] = buffer1[length-3];
         }
      }
      else if (current==1)
      {
         outbuffer[0]        = buffer2[0];
         outbuffer[length-1] = buffer2[length-1];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer2[1];
            outbuffer[length-2] = buffer2[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer2[2];
            outbuffer[length-3] = buffer2[length-3];
         }
      }
      else if (current==2)
      {
         outbuffer[0]        = buffer3[0];
         outbuffer[length-1] = buffer3[length-1];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer3[1];
            outbuffer[length-2] = buffer3[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer3[2];
            outbuffer[length-3] = buffer3[length-3];
         }
      }
      else if (current==3)
      {
         outbuffer[0]        = buffer4[0];
         outbuffer[length-1] = buffer4[length-1];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer4[1];
            outbuffer[length-2] = buffer4[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer4[2];
            outbuffer[length-3] = buffer4[length-3];
         }
      }
      else if (current == 4)
      {
         outbuffer[0]        = buffer5[0];
         outbuffer[length-1] = buffer5[length-1];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer5[1];
            outbuffer[length-2] = buffer5[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer5[2];
            outbuffer[length-3] = buffer5[length-3];
         }
      }
      else if (current == 5)
      {
         outbuffer[0]        = buffer6[0];
         outbuffer[length-1] = buffer6[length-1];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer6[1];
            outbuffer[length-2] = buffer6[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer6[2];
            outbuffer[length-3] = buffer6[length-3];
         }
      }
      else
      {
         outbuffer[0]        = buffer7[0];
         outbuffer[length-1] = buffer7[length-1];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer7[1];
            outbuffer[length-2] = buffer7[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer7[2];
            outbuffer[length-3] = buffer7[length-3];
         }
      }

      /* Process each pixel in the line, skipping first & last pixel. */
      /* Calculate the average and store into outbuffer */

      for (j=mask_size/2; j<length-mask_size/2; j++)
      {
         if (mask_size == 3)
         {
            sum = 0;
            sum += buffer1[j-1] + buffer1[j] + buffer1[j+1];
            sum += buffer2[j-1] + buffer2[j] + buffer2[j+1];
            sum += buffer3[j-1] + buffer3[j] + buffer3[j+1];
            mean = (sum+4)/9;
         }
         else if (mask_size == 5)
         {
            sum = 0;
            sum += buffer1[j-2]+buffer1[j-1]+buffer1[j]+buffer1[j+1]+buffer1[j+2];
            sum += buffer2[j-2]+buffer2[j-1]+buffer2[j]+buffer2[j+1]+buffer2[j+2];
            sum += buffer3[j-2]+buffer3[j-1]+buffer3[j]+buffer3[j+1]+buffer3[j+2];
            sum += buffer4[j-2]+buffer4[j-1]+buffer4[j]+buffer4[j+1]+buffer4[j+2];
            sum += buffer5[j-2]+buffer5[j-1]+buffer5[j]+buffer5[j+1]+buffer5[j+2];
            mean = (sum+12)/25;
         }
         else
         {
            sum = 0;
            sum += buffer1[j-3]+buffer1[j-2]+buffer1[j-1]+buffer1[j]+
                   buffer1[j+1]+buffer1[j+2]+buffer1[j+3];
            sum += buffer2[j-3]+buffer2[j-2]+buffer2[j-1]+buffer2[j]+
                   buffer2[j+1]+buffer2[j+2]+buffer2[j+3];
            sum += buffer3[j-3]+buffer3[j-2]+buffer3[j-1]+buffer3[j]+
                   buffer3[j+1]+buffer3[j+2]+buffer3[j+3];
            sum += buffer4[j-3]+buffer4[j-2]+buffer4[j-1]+buffer4[j]+
                   buffer4[j+1]+buffer4[j+2]+buffer4[j+3];
            sum += buffer5[j-3]+buffer5[j-2]+buffer5[j-1]+buffer5[j]+
                   buffer5[j+1]+buffer5[j+2]+buffer5[j+3];
            sum += buffer6[j-3]+buffer6[j-2]+buffer6[j-1]+buffer6[j]+
                   buffer6[j+1]+buffer6[j+2]+buffer6[j+3];
            sum += buffer7[j-3]+buffer7[j-2]+buffer7[j-1]+buffer7[j]+
                   buffer7[j+1]+buffer7[j+2]+buffer7[j+3];
            mean = (sum+24)/49;
         }
         outbuffer[j] = mean;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(outbuffer,i,sample,length-1);
      if (RefreshLines > 0)
         PutRefresh(outbuffer,i,sample,length);

      /* Get next line and store in appropriate buffer */

      if (i < height+line-mask_size/2-1)
      {
         if (new==0)
            GetLine(buffer1,i+mask_size/2+1,sample,length);
         else if (new==1)
            GetLine(buffer2,i+mask_size/2+1,sample,length);
         else if (new==2)
            GetLine(buffer3,i+mask_size/2+1,sample,length);
         else if (new==3)
            GetLine(buffer4,i+mask_size/2+1,sample,length);
         else if (new==4)
            GetLine(buffer5,i+mask_size/2+1,sample,length);
         else if (new==5)
            GetLine(buffer6,i+mask_size/2+1,sample,length);
         else
            GetLine(buffer7,i+mask_size/2+1,sample,length);
      }
      new++;
      if (new==mask_size) new=0;
      current++;
      if (current==mask_size) current = 0;
   }
}

/*************************************************************************/
/* Edge_Filter                                                           */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/* This routine will apply an edge enhancement filter to the image       */
/* using a 3x3 mask.  There are four variations of this filter.  Integer */
/* math is used to speed up the process.                                 */
/*************************************************************************/

int Edge_Filter(void)
{
   int i,j,k;
   int current=1;
   int new=0;
   int result;
   unsigned int result1;
   int sum;
   unsigned int sum1;
   long offset;
   unsigned char *ptr;
   int  edge;
   int flag;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int factor;

   GetKeywordInteger (CommandString, "EDG", 0, &edge, &flag);
   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if ((edge < 1) || (edge > 3)) edge = 0;
   if (edge <= 1)
      factor = 10;
   else
      factor = 9;

   /* Apply filter to just the image or the entire screen */

   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }


   /* Read in the first three lines of the image */

   GetLine(buffer1,line,sample,length);
   GetLine(buffer2,line+1,sample,length);
   GetLine(buffer3,line+2,sample,length);

   /* Loop through the image line-by-line, but skip the first & last line */

   for (i=line+1; i<=line+height-2; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Store first and last pixel into the outbuffer */
      if (current==0)
      {
         outbuffer[0]        = buffer1[0];
         outbuffer[length-1] = buffer1[length-1];
         ptr = &buffer1[1];
      }
      else if (current==1)
      {
         outbuffer[0]        = buffer2[0];
         outbuffer[length-1] = buffer2[length-1];
         ptr = &buffer2[1];
      }
      else
      {
         outbuffer[0]        = buffer3[0];
         outbuffer[length-1] = buffer3[length-1];
         ptr = &buffer3[1];
      }

      if ((edge == 0) || (edge == 2))
      {
         /* Process each pixel in the line, skipping first & last pixel */
         for (j=1; j<length-1; j++, ptr++)
         {
            sum = 0;
            sum += buffer1[j-1] + buffer1[j] + buffer1[j+1];
            sum += buffer2[j-1] + buffer2[j] + buffer2[j+1];
            sum += buffer3[j-1] + buffer3[j] + buffer3[j+1];
            result = *ptr * factor - sum;
            if (result < 0) result = *ptr;
            if (result >= numDN) result = *ptr;
            outbuffer[j] = result;
         }
      }
      else
      {
         /* Process each pixel in the line, skipping first & last pixel */
         for (j=1; j<length-1; j++, ptr++)
         {
            sum1 = 0;
            sum1 += buffer1[j-1] + buffer1[j] + buffer1[j+1];
            sum1 += buffer2[j-1] + buffer2[j] + buffer2[j+1];
            sum1 += buffer3[j-1] + buffer3[j] + buffer3[j+1];
            result1 = *ptr * factor - sum1;
            if (result1 < 0) result1 = 0;
            if (result1 >= numDN) result1 = numDN-1;
            outbuffer[j] = result1;
         }
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(outbuffer,i,sample,length-1);
      if (RefreshLines > 0)
         PutRefresh(outbuffer,i,sample,length);

      /* Get next line and store in appropriate buffer */

      if (i < height+line-2)
      {
         if (new==0)
            GetLine(buffer1,i+2,sample,length);
         else if (new==1)
            GetLine(buffer2,i+2,sample,length);
         else
            GetLine(buffer3,i+2,sample,length);
      }
      new++;
      if (new==3) new=0;
      current++;
      if (current==3) current = 0;
   }
}

/*************************************************************************/
/* Convolution_Filter                                                    */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/* This routine will apply a  convolution filter to the image which      */
/* has the effect of detecting the edges.  A 3x3 mask is used.  Integer  */
/* math is used to speed up the process.  The following weighted mask    */
/* is used:                                                              */
/*               -1  -1  -1                                              */
/*               -1   8  -1                                              */
/*               -1  -1  -1                                              */
/*                                                                       */
/* There are two variations of this filter used here.                    */
/*************************************************************************/

int Convolution_Filter(void)
{
   int i,j,k;
   int current=1;
   int new=0;
   int sum;
   long offset;
   int  edge, flag;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;

   GetKeywordInteger (CommandString, "CON", 0, &edge, &flag);
   if (edge != 1) edge = 0;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Read in the first three lines of the image */

   GetLine(buffer1,line,sample,length);
   GetLine(buffer2,line+1,sample,length);
   GetLine(buffer3,line+2,sample,length);

   /* Loop through the image line-by-line, but skip the first & last line */

   for (i=line+1; i<=line+height-2; i++)
   {
      if (kbhit())       /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Store first and last pixel into the outbuffer */

      if (current==0)
      {
         outbuffer[0]        = buffer1[0];
         outbuffer[length-1] = buffer1[length-1];
      }
      else if (current==1)
      {
         outbuffer[0]        = buffer2[0];
         outbuffer[length-1] = buffer2[length-1];
      }
      else
      {
         outbuffer[0]        = buffer3[0];
         outbuffer[length-1] = buffer2[length-1];
      }

      /* Process each pixel in the line, skipping first & last pixel */

      for (j=1; j<length-1; j++)
      {
         sum = 0;
         if (current==0)
         {
            sum += 8 * buffer1[j] - buffer1[j-1] - buffer1[j+1];
            sum += -buffer2[j-1]  - buffer2[j]   - buffer2[j+1];
            sum += -buffer3[j-1]  - buffer3[j]   - buffer3[j+1];
         }
         else if (current==1)
         {
            sum += 8 * buffer2[j] - buffer2[j-1] - buffer2[j+1];
            sum += -buffer1[j-1]  - buffer1[j]   - buffer1[j+1];
            sum += -buffer3[j-1]  - buffer3[j]   - buffer3[j+1];
         }
         else
         {
            sum += 8 * buffer3[j] - buffer3[j-1] - buffer3[j+1];
            sum += -buffer2[j-1]  - buffer2[j]   - buffer2[j+1];
            sum += -buffer1[j-1]  - buffer1[j]   - buffer1[j+1];
         }
         if (edge==0)
         {
            if (sum < 0) sum = 0;
            if (sum >= numDN) sum = numDN-1;
         }
         outbuffer[j] = sum;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(outbuffer,i,sample,length-1);
      if (RefreshLines > 0)
         PutRefresh(outbuffer,i,sample,length);

      /* Get next line and store in appropriate buffer */

      if (i < height+line-2)
      {
         if (new==0)
            GetLine(buffer1,i+2,sample,length);
         else if (new==1)
            GetLine(buffer2,i+2,sample,length);
         else
            GetLine(buffer3,i+2,sample,length);
      }
      new++;
      if (new==3) new=0;
      current++;
      if (current==3) current = 0;
   }
}
/*************************************************************************/
/* Median_Filter                                                         */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/* This routine will apply a median filter to the image which has the    */
/* effect of smoothing out the image.  A 3x3 mask is used where the      */
/* middle pixel is replaced by the middle of the nine pixels after they  */
/* have been sorted.  The sort used is a simple bubble sort which is     */
/* terminated halfway through the sort since the median value will be    */
/* found by then.                                                        */
/*************************************************************************/

int Median_Filter(void)
{
   #define ASIZE 9
   int i,j,k,l;
   int current=1;
   int new=0;
   long offset;
   unsigned array[ASIZE];
   int flag;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   unsigned temp;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Read in the first three lines of the image */

   GetLine(buffer1,line,sample,length);
   GetLine(buffer2,line+1,sample,length);
   GetLine(buffer3,line+2,sample,length);

   /* Loop through the image line-by-line, but skip the first & last line */

   for (i=line+1; i<=line+height-2; i++)
   {
      if (kbhit())       /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Store first and last pixel into the outbuffer */
      if (current==0)
      {
         outbuffer[0]        = buffer1[0];
         outbuffer[length-1] = buffer1[length-1];
      }
      else if (current==1)
      {
         outbuffer[0]        = buffer2[0];
         outbuffer[length-1] = buffer2[length-1];
      }
      else
      {
         outbuffer[0]        = buffer3[0];
         outbuffer[length-1] = buffer2[length-1];
      }

      /* Process each pixel in the line, skipping first & last pixel */

      for (j=1; j<length-1; j++)
      {
         /* Store the nine pixels into the array */
         array[0] = buffer1[j-1];
         array[1] = buffer1[j];
         array[2] = buffer1[j+1];
         array[3] = buffer2[j-1];
         array[4] = buffer2[j];
         array[5] = buffer2[j+1];
         array[6] = buffer3[j-1];
         array[7] = buffer3[j];
         array[8] = buffer3[j+1];

         /* sort the array */

         insert(9, array);
         outbuffer[j] = array[4];   /* Store middle pixel */
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(outbuffer,i,sample,length-1);
      if (RefreshLines > 0)
         PutRefresh(outbuffer,i,sample,length);

      /* Get next line and store in appropriate buffer */

      if (i < height+line-2)
      {
         if (new==0)
            GetLine(buffer1,i+2,sample,length);
         else if (new==1)
            GetLine(buffer2,i+2,sample,length);
         else
            GetLine(buffer3,i+2,sample,length);
      }
      new++;
      if (new==3) new=0;
      current++;
      if (current==3) current = 0;
   }
}

/*************************************************************************/
/* Dark_Filter                                                           */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/* Not really a filter, but this routine will either darken or brighten  */
/* an images a specified amount.  If no value is specified then the      */
/* amount of darkness/brightness willl be 10% of the max color display   */
/* (ie. 2 for 16 color, 25 for 256 color).                               */
/*************************************************************************/

int Dark_Filter(int dark_bright_flag)
{
   int i,j;
   int  amount, flag;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int temp;

   /* Check to see if darken or brighten was chosen */

   if (dark_bright_flag == 0)
      GetKeywordInteger (CommandString, "BRI", 0, &amount, &flag);
   else
      GetKeywordInteger (CommandString, "DAR", 0, &amount, &flag);
   if (amount<=0)
      amount = (numDN + 5)/10;
   if (dark_bright_flag == 1)
      amount = -amount;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Loop through the image line-by-line */

   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      for (j=0; j<length; j++)
      {
         temp = buffer1[j] + amount;
         if (temp < 0) temp = 0;
         if (temp >= numDN) temp = numDN-1;
         buffer1[j] = temp;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(buffer1,i,sample,length);
      if (RefreshLines > 0)
         PutRefresh(buffer1,i,sample,length);
   }
}

/*************************************************************************/
/* Smear_Filter                                                          */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/*************************************************************************/

int Smear_Filter(void)
{
   int i,j;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int temp;
   int current=0;
   int divisor=8;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
      divisor = divisor * Image_Subsample;
   }

   /* Loop through the image line-by-line */

   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      GetLine(buffer2,i,sample,length);
      current = i;
      for (j=0; j<length-1; j++)
      {
         temp = i - line - buffer1[j]/divisor;
         if (temp <= 0) temp = 1;
         temp += line;
         if (temp != current)
         {
            DisplayLine(buffer2,current,sample,length);
            if (RefreshLines > 0)
               PutRefresh(buffer2,current,sample,length);
            GetLine(buffer2,temp,sample,length);
            current = temp;
         }
         buffer2[j] = buffer1[j];
      }

      DisplayLine(buffer2,current,sample,length);
      if (RefreshLines > 0)
         PutRefresh(buffer2,current,sample,length);
   }
}
/*************************************************************************/
/* DoSlant                                                               */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/* This routine will shift each line of the image over one pixel causing */
/* a slant effect.  The image can be slanted to either the left or right */
/* direction.                                                            */
/*************************************************************************/

int DoSlant(void)
{
   int i,j,k;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int slant_right;

   /* Slant left or right */

   GetKeywordString (CommandString, "LEF", "",dummy, &slant_right);
   if (slant_right != -1)
      slant_right = FALSE;
   else
      slant_right = TRUE;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Loop through the image line-by-line */

   k=1;
   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      for (j=0; j<length; j++)
      {
         if (slant_right)
            buffer2[j] = buffer1[(j-k+length+1) % length];
         else
            buffer2[j] = buffer1[(j+k-1) % length];
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(buffer2,i,sample,length);
      if (RefreshLines > 0)
         PutRefresh(buffer2,i,sample,length);
      k++;
   }
}

/*************************************************************************/
/* Baalke_Filter                                                         */
/*                                                                       */
/* Written by Ron Baalke - 05/91                                         */
/*                                                                       */
/*************************************************************************/

int Baalke_Filter(void)
{
   int i,j,k,l;
   int current=1;
   int new=0;
   int result;
   int sum;
   unsigned char *ptr;
   int flag;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int temp;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   temp = length * 20;

   /* Read in the first three lines of the image */

   GetLine(buffer1,line,sample,length);
   GetLine(outbuffer,line+1,sample,length);
   for (i=0; i<length; i++)
      buffer2[i] = outbuffer[(i-3+length) % length];
   GetLine(outbuffer,line+2,sample,length);
   for (i=0; i<length; i++)
      buffer3[i] = outbuffer[(i-6+length) % length];
   /* Loop through the image line-by-line, but skip the first & last line */

   l = 2;
   for (i=line+1; i<=line+height-2; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Store first and last pixel into the outbuffer */
      if (current==0)
      {
         outbuffer[0]        = buffer1[0];
         outbuffer[length-1] = buffer1[length-1];
         ptr = &buffer1[1];
      }
      else if (current==1)
      {
         outbuffer[0]        = buffer2[0];
         outbuffer[length-1] = buffer2[length-1];
         ptr = &buffer2[1];
      }
      else
      {
         outbuffer[0]        = buffer3[0];
         outbuffer[length-1] = buffer3[length-1];
         ptr = &buffer3[1];
      }

      /* Process each pixel in the line, skipping first & last pixel */
      for (j=1; j<length-1; j++, ptr++)
      {
         sum = 0;
         sum += buffer1[j-1] + buffer1[j] + buffer1[j+1];
         sum += buffer2[j-1] + buffer2[j] + buffer2[j+1];
         sum += buffer3[j-1] + buffer3[j] + buffer3[j+1];
         result = *ptr * 10 - sum;
         if (result < 0) result = *ptr;
         if (result >= numDN) result = *ptr;
         outbuffer[j] = result;
      }

      for (k=0; k<length; k++)
         buffer4[k] = outbuffer[(k+(l-1)*3) % length];
      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(buffer4,i,sample,length-1);
      if (RefreshLines > 0)
         PutRefresh(buffer4,i,sample,length);

      /* Get next line and store in appropriate buffer */

      if (i < height+line-2)
      {
         if (new==0)
         {
            GetLine(outbuffer,i+2,sample,length);
            for (k=0; k<length; k++)
               buffer1[k] = outbuffer[(k-(l+1)*3+temp) % length];
         }
         else if (new==1)
         {
            GetLine(outbuffer,i+2,sample,length);
            for (k=0; k<length; k++)
               buffer2[k] = outbuffer[(k-(l+1)*3+temp) % length];
         }
         else
         {
            GetLine(outbuffer,i+2,sample,length);
            for (k=0; k<length; k++)
               buffer3[k] = outbuffer[(k-(l+1)*3+temp) % length];
         }
      }
      new++;
      if (new==3) new=0;
      current++;
      if (current==3) current = 0;
      l++;
   }
}

/*************************************************************************/
/* Unsharp_Filter                                                        */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* This routine will sharpen the image by subtracting 3 times the        */
/* original image from 2 times the blurred image.  A 5x5 mean filter     */
/* is used for the blurred portion.                                      */
/*************************************************************************/
int Unsharp_Filter(void)
{
   int i,j,k;
   int current;
   int new=0;
   int mean;
   unsigned int sum;
   long offset;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int mask_size;
   int flag;


   mask_size = 5;
   current = 2;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Read in the first 5 lines of the image */

   GetLine(buffer1,line,sample,length);
   GetLine(buffer2,line+1,sample,length);
   GetLine(buffer3,line+2,sample,length);
   GetLine(buffer4,line+3,sample,length);
   GetLine(buffer5,line+4,sample,length);

   /* Loop through the image line-by-line, but skip the first & last line */

   for (i=line+mask_size/2; i<=line+height- mask_size/2 -1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Store first and last pixel into the outbuffer */

      if (current==0)
      {
         outbuffer[0]        = buffer1[0];
         outbuffer[length-1] = buffer1[length-1];
         if (mask_size == 5)
         {
            outbuffer[1]        = buffer1[1];
            outbuffer[length-2] = buffer1[length-2];
         }
      }
      else if (current==1)
      {
         outbuffer[0]        = buffer2[0];
         outbuffer[length-1] = buffer2[length-1];
         if (mask_size == 5)
         {
            outbuffer[1]        = buffer2[1];
            outbuffer[length-2] = buffer2[length-2];
         }
      }
      else if (current==2)
      {
         outbuffer[0]        = buffer3[0];
         outbuffer[length-1] = buffer3[length-1];
         if (mask_size == 5)
         {
            outbuffer[1]        = buffer3[1];
            outbuffer[length-2] = buffer3[length-2];
         }
      }
      else if (current==3)
      {
         outbuffer[0]        = buffer4[0];
         outbuffer[length-1] = buffer4[length-1];
         if (mask_size == 5)
         {
            outbuffer[1]        = buffer4[1];
            outbuffer[length-2] = buffer4[length-2];
         }
      }
      else
      {
         outbuffer[0]        = buffer5[0];
         outbuffer[length-1] = buffer5[length-1];
         if (mask_size == 5)
         {
            outbuffer[1]        = buffer5[1];
            outbuffer[length-2] = buffer5[length-2];
         }
      }

      /* Process each pixel in the line, skipping first & last pixel. */
      /* Calculate the average and store into outbuffer */

      for (j=mask_size/2; j<length-mask_size/2; j++)
      {
         sum = 0;
         sum += buffer1[j-2]+buffer1[j-1]+buffer1[j]+buffer1[j+1]+buffer1[j+2];
         sum += buffer2[j-2]+buffer2[j-1]+buffer2[j]+buffer2[j+1]+buffer2[j+2];
         sum += buffer3[j-2]+buffer3[j-1]+buffer3[j]+buffer3[j+1]+buffer3[j+2];
         sum += buffer4[j-2]+buffer4[j-1]+buffer4[j]+buffer4[j+1]+buffer4[j+2];
         sum += buffer5[j-2]+buffer5[j-1]+buffer5[j]+buffer5[j+1]+buffer5[j+2];
         mean = (sum+12)/25;
         if (current == 0)
            mean = 3 * buffer1[j] - 2 * mean;
         else if (current == 1)
            mean = 3 * buffer2[j] - 2 * mean;
         else if (current == 2)
            mean = 3 * buffer3[j] - 2 * mean;
         else if (current == 3)
            mean = 3 * buffer4[j] - 2 * mean;
         else
            mean = 3 * buffer5[j] - 2 * mean;
         if (mean < 0) mean = 0;
         if (mean >= numDN) mean = numDN-1;

         outbuffer[j] = mean;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(outbuffer,i,sample,length-1);
      if (RefreshLines > 0)
         PutRefresh(outbuffer,i,sample,length);

      /* Get next line and store in appropriate buffer */

      if (i < height+line-mask_size/2-1)
      {
         if (new==0)
            GetLine(buffer1,i+mask_size/2+1,sample,length);
         else if (new==1)
            GetLine(buffer2,i+mask_size/2+1,sample,length);
         else if (new==2)
            GetLine(buffer3,i+mask_size/2+1,sample,length);
         else if (new==3)
            GetLine(buffer4,i+mask_size/2+1,sample,length);
         else
            GetLine(buffer5,i+mask_size/2+1,sample,length);
      }
      new++;
      if (new==mask_size) new=0;
      current++;
      if (current==mask_size) current = 0;
   }
}

/*************************************************************************/
/* Logarithmic_Filter                                                    */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* Replaces each pixel with its log multiplied by a factor.  The factor  */
/* is user specified (default=40).  Works well with dark images or       */
/* stars and galaxies.  To keep floating point calculations to a minimum,*/
/* an integer log table is created first, and then integer arithmetic is */
/* used wherever possible                                                */
/*************************************************************************/

int Logarithmic_Filter(void)
{
   int i,j;
   int  amount, flag;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int temp;
   long logtable[256];
   long amount1;

   GetKeywordInteger (CommandString, "LOG", 40, &amount, &flag);

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Build log table */

   for (i=0; i<256; i++)
   {
      logtable[i] = (long)log((double)i + 1.0) * 1000;
   }
   amount1 = 1000 * amount;
   /* Loop through the image line-by-line */

   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      for (j=0; j<length; j++)
      {
         temp = (int) ((logtable[buffer1[j]] * amount + 500)/1000);
         if (temp < 0) temp = 0;
         if (temp >= numDN) temp = numDN-1;
         buffer1[j] = temp;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(buffer1,i,sample,length);
      if (RefreshLines > 0)
         PutRefresh(buffer1,i,sample,length);
   }
}

/*************************************************************************/
/* Power_Filter                                                          */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* Replaces each pixel with the following equation:                      */
/*                                                                       */
/*     new pixel = numDN * (old/numDN) ^ power                           */
/*                                                                       */
/* where power is specified by the user (default = 2.0)                  */
/* To keep floating point calculations to a minimum, a power table is    */
/* created first, and then integer arithmetic used wherever possible     */
/*************************************************************************/

int Power_Filter(void)
{
   int i,j;
   int  flag;
   float ctemp;
   double amount;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int temp;
   int powertable[256];

   GetKeywordReal (CommandString, "POW", 2.0, &ctemp, &flag);
   amount = (double) ctemp;

   if (amount == 1.0) return(0);

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Build power table */

   for (i=0; i<256; i++)
   {
      powertable[i] = (int) (pow((double)i/(double)numDN, amount) *
                            (double)numDN + .5);
   }
   /* Loop through the image line-by-line */

   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      for (j=0; j<length; j++)
      {
         temp = powertable[buffer1[j]];
         if (temp < 0) temp = 0;
         if (temp >= numDN) temp = numDN-1;
         buffer1[j] = temp;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(buffer1,i,sample,length);
      if (RefreshLines > 0)
         PutRefresh(buffer1,i,sample,length);
   }
}
/*************************************************************************/
/* Stair_Filter                                                          */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* Replaces each pixel with its log multiplied by a factor.  The factor  */
/* is user specified (default=40).  Works well with dark images or       */
/* stars and galaxies.  To keep floating point calculations to a minimum,*/
/* an integer log table is created first, and then integer arithmetic is */
/* used wherever possible                                                */
/*************************************************************************/

int Stair_Filter(void)
{
   int i,j;
   int  steps, stepsize, flag;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int temp;
   int stairtable[256];

   GetKeywordInteger (CommandString, "STA", 8, &steps, &flag);
   if (steps <= 1) steps = 2;
   stepsize = numDN/steps;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Build stair table */

   for (i=0; i<256; i++)
      stairtable[i] = i;
   for (i=0; i<256; i++)
   {
      stairtable[i] = stepsize * (i/stepsize);
   }

   /* Loop through the image line-by-line */

   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      for (j=0; j<length; j++)
      {
         temp = stairtable[buffer1[j]];
         if (temp < 0) temp = 0;
         if (temp >= numDN) temp = numDN-1;
         buffer1[j] = temp;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(buffer1,i,sample,length);
      if (RefreshLines > 0)
         PutRefresh(buffer1,i,sample,length);
   }
}

/*************************************************************************/
/* DoMask                                                                */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* This command will mask out either the lower or upper pixels values    */
/* of an image.                                                          */
/*************************************************************************/

int DoMask(void)
{
   int i,j;
   int lo_flag, hi_flag, with_flag;
   int old_value, amount;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int temp;

   /* Check to see if darken or brighten was chosen */

   GetKeywordInteger (CommandString, "LO", 0, &old_value, &lo_flag);
   if (lo_flag == -1)
      GetKeywordInteger (CommandString, "HI", 0, &old_value, &hi_flag);
   GetKeywordInteger (CommandString, "WITH", 0, &amount, &with_flag);

   if ((lo_flag == -1) && (hi_flag == -1))
   {
      StatusLine(0,"Invalid Parameters");
      return(1);
   }

   if ((lo_flag != -1) && (with_flag == -1))
      amount = 0;
   else if ((hi_flag != -1) && (with_flag == -1))
      amount = numDN - 1;

   if (amount < 0) amount = 0;
   if (amount >= numDN) amount = numDN - 1;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Loop through the image line-by-line */

   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      for (j=0; j<length; j++)
      {
         temp = buffer1[j];
         if (lo_flag != -1)
         {
            if (temp <= old_value)
               buffer1[j] = amount;
         }
         else if (temp >= old_value)
            buffer1[j] = amount;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(buffer1,i,sample,length);
      if (RefreshLines > 0)
         PutRefresh(buffer1,i,sample,length);
   }
}

/*************************************************************************/
/* HighPass_Filter                                                       */
/*                                                                       */
/* Written by Ron Baalke - 02/92                                         */
/*                                                                       */
/* This routine will apply a high pass filter to the image.  The middle  */
/* pixel of the mask is subtracted from the mean of the mask, and then   */
/* 1/2 of the number of color is added, and then a stretch is performed. */
/*************************************************************************/
int HighPass_Filter(void)
{
   int i,j,k;
   int current;
   int new=0;
   int mean;
   int sum;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   int mask_size;
   int flag;
   int adj;
   unsigned char *ptr;

   /* Apply filter to just the image or the entire screen */

   GetKeywordInteger (CommandString, "HIG",0, &mask_size, &flag);
   if ((mask_size < 1) || (mask_size > 2)) mask_size = 0;
   if (mask_size == 0)
   {
      mask_size = 3;
      current = 1;
   }
   else if (mask_size == 1)
   {
      mask_size = 5;
      current = 2;
   }
   else
   {
      mask_size = 7;
      current = 3;
   }

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Read in the first three lines of the image */

   GetLine(buffer1,line,sample,length);
   GetLine(buffer2,line+1,sample,length);
   GetLine(buffer3,line+2,sample,length);

   /* If 5x5 mask then read in two more lines */

   if (mask_size == 5)
   {
      GetLine(buffer4,line+3,sample,length);
      GetLine(buffer5,line+4,sample,length);
   }
   else if (mask_size == 7)
   {
      GetLine(buffer6,line+5,sample,length);
      GetLine(buffer7,line+6,sample,length);
   }

   /* Loop through the image line-by-line, but skip the first & last line */

   for (i=line+mask_size/2; i<=line+height- mask_size/2 -1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Store first and last pixel into the outbuffer */

      if (current==0)
      {
         outbuffer[0]        = buffer1[0];
         outbuffer[length-1] = buffer1[length-1];
         ptr = &buffer1[mask_size/2];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer1[1];
            outbuffer[length-2] = buffer1[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer1[2];
            outbuffer[length-3] = buffer1[length-3];
         }
      }
      else if (current==1)
      {
         outbuffer[0]        = buffer2[0];
         outbuffer[length-1] = buffer2[length-1];
         ptr = &buffer2[mask_size/2];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer2[1];
            outbuffer[length-2] = buffer2[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer2[2];
            outbuffer[length-3] = buffer2[length-3];
         }
      }
      else if (current==2)
      {
         outbuffer[0]        = buffer3[0];
         outbuffer[length-1] = buffer3[length-1];
         ptr = &buffer3[mask_size/2];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer3[1];
            outbuffer[length-2] = buffer3[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer3[2];
            outbuffer[length-3] = buffer3[length-3];
         }
      }
      else if (current==3)
      {
         outbuffer[0]        = buffer4[0];
         outbuffer[length-1] = buffer4[length-1];
         ptr = &buffer4[mask_size/2];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer4[1];
            outbuffer[length-2] = buffer4[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer4[2];
            outbuffer[length-3] = buffer4[length-3];
         }
      }
      else if (current == 4)
      {
         outbuffer[0]        = buffer5[0];
         outbuffer[length-1] = buffer5[length-1];
         ptr = &buffer5[mask_size/2];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer5[1];
            outbuffer[length-2] = buffer5[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer5[2];
            outbuffer[length-3] = buffer5[length-3];
         }
      }
      else if (current == 5)
      {
         outbuffer[0]        = buffer6[0];
         outbuffer[length-1] = buffer6[length-1];
         ptr = &buffer6[mask_size/2];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer6[1];
            outbuffer[length-2] = buffer6[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer6[2];
            outbuffer[length-3] = buffer6[length-3];
         }
      }
      else
      {
         outbuffer[0]        = buffer7[0];
         outbuffer[length-1] = buffer7[length-1];
         ptr = &buffer7[mask_size/2];
         if (mask_size >= 5)
         {
            outbuffer[1]        = buffer7[1];
            outbuffer[length-2] = buffer7[length-2];
         }
         if (mask_size >= 7)
         {
            outbuffer[2]        = buffer7[2];
            outbuffer[length-3] = buffer7[length-3];
         }
      }

      /* Process each pixel in the line, skipping first & last pixel. */
      /* Perform high pass calculations and store into outbuffer */
      adj = (numDN-1)/2;
      for (j=mask_size/2; j<length-mask_size/2; j++, ptr++)
      {
         if (mask_size == 3)
         {
            sum = 0;
            sum += buffer1[j-1] + buffer1[j] + buffer1[j+1];
            sum += buffer2[j-1] + buffer2[j] + buffer2[j+1];
            sum += buffer3[j-1] + buffer3[j] + buffer3[j+1];
            mean = (sum+4)/9;
            mean += adj - *ptr;
         }
         else if (mask_size == 5)
         {
            sum = 0;
            sum += buffer1[j-2]+buffer1[j-1]+buffer1[j]+buffer1[j+1]+buffer1[j+2];
            sum += buffer2[j-2]+buffer2[j-1]+buffer2[j]+buffer2[j+1]+buffer2[j+2];
            sum += buffer3[j-2]+buffer3[j-1]+buffer3[j]+buffer3[j+1]+buffer3[j+2];
            sum += buffer4[j-2]+buffer4[j-1]+buffer4[j]+buffer4[j+1]+buffer4[j+2];
            sum += buffer5[j-2]+buffer5[j-1]+buffer5[j]+buffer5[j+1]+buffer5[j+2];
            mean = (sum+12)/25;
            mean += adj - *ptr;
         }
         else
         {
            sum = 0;
            sum += buffer1[j-3]+buffer1[j-2]+buffer1[j-1]+buffer1[j]+
                   buffer1[j+1]+buffer1[j+2]+buffer1[j+3];
            sum += buffer2[j-3]+buffer2[j-2]+buffer2[j-1]+buffer2[j]+
                   buffer2[j+1]+buffer2[j+2]+buffer2[j+3];
            sum += buffer3[j-3]+buffer3[j-2]+buffer3[j-1]+buffer3[j]+
                   buffer3[j+1]+buffer3[j+2]+buffer3[j+3];
            sum += buffer4[j-3]+buffer4[j-2]+buffer4[j-1]+buffer4[j]+
                   buffer4[j+1]+buffer4[j+2]+buffer4[j+3];
            sum += buffer5[j-3]+buffer5[j-2]+buffer5[j-1]+buffer5[j]+
                   buffer5[j+1]+buffer5[j+2]+buffer5[j+3];
            sum += buffer6[j-3]+buffer6[j-2]+buffer6[j-1]+buffer6[j]+
                   buffer6[j+1]+buffer6[j+2]+buffer6[j+3];
            sum += buffer7[j-3]+buffer7[j-2]+buffer7[j-1]+buffer7[j]+
                   buffer7[j+1]+buffer7[j+2]+buffer7[j+3];
            mean = (sum+24)/49;
            mean += adj - *ptr;

         }
         if (mean < 0) mean = 0;
         if (mean >= numDN) mean = numDN-1;
         if (numDN == 256)
         {
            if (mean < 120) mean = 0;
            if (mean > 134) mean = numDN-1;
            mean = (134-mean) * 17;
         }
         outbuffer[j] = mean;
      }

      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(outbuffer,i,sample,length-1);
      if (RefreshLines > 0)
         PutRefresh(outbuffer,i,sample,length);

      /* Get next line and store in appropriate buffer */

      if (i < height+line-mask_size/2-1)
      {
         if (new==0)
            GetLine(buffer1,i+mask_size/2+1,sample,length);
         else if (new==1)
            GetLine(buffer2,i+mask_size/2+1,sample,length);
         else if (new==2)
            GetLine(buffer3,i+mask_size/2+1,sample,length);
         else if (new==3)
            GetLine(buffer4,i+mask_size/2+1,sample,length);
         else if (new==4)
            GetLine(buffer5,i+mask_size/2+1,sample,length);
         else if (new==5)
            GetLine(buffer6,i+mask_size/2+1,sample,length);
         else
            GetLine(buffer7,i+mask_size/2+1,sample,length);
      }
      new++;
      if (new==mask_size) new=0;
      current++;
      if (current==mask_size) current = 0;
   }
}
/*************************************************************************/
/* Histogram_Equalization                                                */
/*                                                                       */
/* Written by Ron Baalke - 02/92                                         */
/*                                                                       */
/*************************************************************************/

int Histogram_Equalization(void)
{
   int i,j;
   int line;
   int sample;
   int height;
   int length;
   char dummy[80];
   int screen;
   long sum;
   long histbuf[256];
   double ftemp;

   /* Apply filter to just the image or the entire screen */

   GetKeywordString (CommandString, "SCR", "",dummy, &screen);
   if (screen != -1)
   {
      line = 1;
      sample = 1;
      height = dispnl;
      length = dispns;
   }
   else
   {
      line   = Image_Line;
      sample = Image_Sample;
      height = Image_Height;
      length = Image_Length;
   }

   /* Build the Histogram */

   for (i=0; i<numDN; i++)
      histbuf[i] = 0;

   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      for (j=0; j<length; j++)
      {
         histbuf[buffer1[j]]++;
      }
   }

   sum = 0;
   for (i=0; i<numDN; i++)
   {
      sum += histbuf[i];
      histbuf[i] = sum;
   }

   ftemp = (double)numDN/(double)sum;
   for (i=0; i<numDN; i++)
   {
      histbuf[i] = (long)(ftemp * (double) histbuf[i]);
      if (histbuf[i] >= numDN) histbuf[i] = numDN-1;
   }

   for (i=line; i<=line+height-1; i++)
   {
      if (kbhit())    /* Allow user to quit anytime */
      {
         getch();
         return(0);
      }

      /* Read in each line and process each pixel in the line */

      GetLine(buffer1,i,sample,length);
      for (j=0; j<length; j++)
      {
         outbuffer[j] = (unsigned char)histbuf[buffer1[j]];
      }


      /* Write the line out onto the screen, and save to refresh buffer */
      /* if available                                                   */

      DisplayLine(outbuffer,i,sample,length);
      if (RefreshLines > 0)
         PutRefresh(outbuffer,i,sample,length);
   }
}
