/**********************************************************************/
/*  Original Copyright (C) 1989, California Institute of Technology   */
/*  U. S. Government Sponsorship under NASA Contract                  */
/*  NAS7-918 is acknowledged.                                         */
/**********************************************************************/

/***  Program: IMDISP

     Interactive Image Display Program

***/

#define __MSC

/* * * * INCLUDE files * * * */

#include <conio.h>
#include <ctype.h>
#include <direct.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <malloc.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mshell.h"
#include "imdef.h"
#include "imdutil.h"
#include "dispio.h"
#include "disputil.h"
#include "display.h"
#include "fileio.h"
#include "fileutil.h"
#include "help.h"
#include "imageio.h"
#include "imagutil.h"
#include "keywutil.h"
#include "labutil.h"
#include "plot.h"
#include "refresh.h"
#include "swap.h"
#include "browse.h"
#include "palette.h"
#include "palutil.h"
#include "textutil.h"
#include "filter.h"
#include "buffer.h"
#include "gif_lib.h"
#include "mem.h"

/* * * * External functions * * * */
extern void pascal far mouses( int *, int *, int *, int * );

/* * * * Function declarations * * * */

int    DoRefresh (void);
int    SetValues (void);
int    DoHistogram (void);
int    DoPalette (void);
int    DoStretch (void);
int    DoCursor (void);
int    DoProfile (void);
int    DoSave (void);
int    DoBatch (void);
int    DoLabel (void);
int    DoMenu (void);
int    DoPosText (void);
int    DoOverlay (void);
int    DoSerial (void);
int    DoPrint (void);
int    DoLog (void);

/* * * * Global Variables * * * */

Color   CT[256];
char   CommandString[80];
int    nl, ns, bitsperpix;
unsigned int labelsize;
long   DNlow, DNhigh;
int    OpenFileFlag, SubsampleFlag;
int    nld, nsd, sl, ss;
int    nldd, nsdd, sldd, ssdd;
int    subsample, zoom;
int    centerline, centersamp;
int    lastsl, lastss;
FILE   *BatchFile[MAXBATCHFILES];
Spectrum baseliner;
Image  ThisImage;

long   Histogram;       /* Location of histogram in file */
long   Palette;         /* Location of palette in file */
char   PaletteFileName[64];
int    Color_Hist;
int    Microsoft;
int    TurnCursorOff;    /* Set cursor funct. on */
int    BatchFlag;        /* To control batch file on input cmd line */
int    BatchFileNumber=-1;
int    DelayFlag = 0;
int    PromptFlag = 1;
int    flowthru;
int    IntoMem;
int    abort_disp ;

int    ByteSwap = FALSE;
char   memmsg[] = "Insufficient memory.";
char   dirMask[30] = "*.*";               /* select all files */
char   palMask[30] = "*.*";               /* palette mask */
char   PalettePath[30] = ".\\PALETTES\\"; /* default palette directory */
char   BrowseName[40] = "C:\\BROWSE.CMD"; /* default browse file name */
char   MenuName[40] = "MENU.MNU";         /* default menu name */
char   DefaultDir[67];                    /* starting path */
unsigned DefaultDrive, NumDrives;         /* starting drive */

FILE   *SelectFile;
int    SelectFlag;       /* To control interactive selection of files */
char   ImageFileName[80];
int    ImageFormat;
char   SelectName[40] = "C:\\IMDISP.SEL"; /* default select file name */
int    use_mouse;              /* mouse flag */
FILE   *LogFile;
int    logflag;
char   CommandList[MAXRECALL][80];
int    CommandPtr;
int    CommandFlag;

char  helpfilename[80];
FILE *helpfile;
char  ProgramName[80];

unsigned char *buffer1;
unsigned char *buffer2;
unsigned char *buffer3;
unsigned char *buffer4;
unsigned char *buffer5;
unsigned char *buffer6;
unsigned char *buffer7;
unsigned char *outbuffer;

/******************** Start of main program ***********************/

int main (int argc, char *argv[])

{
union    REGS  inregs, outregs;
struct   SREGS segregs;
unsigned int   cur_dseg_var;
char           execute[80];
int            command, len, i, j, k, flag, clearflag;
char           status[128], temp[136];
char           DosCommand[72];
char           *result, *comspec, *imbrowse, SwapName[40];
int            swap_return, sys_return, save_refresh;
unsigned char  exec_return;
int            comment;

signal(SIGINT,ctrl_c);

StartUp (argc, argv);

if ( (_osmajor > 3) || ((_osmajor == 3) && (_osminor >=3)) )
{
    strcpy( ProgramName, argv[0]);
}
else
{
    strcpy( ProgramName, "");
}

strcpy( helpfilename, "IMDISP.HLP");
strcpy( SwapName, "C:\\IMDISWAP.FIL");


use_mouse = 0;                         /* Test mouse function */
mouses( &use_mouse, &i, &j, &k );      /* Set mouse flag */

logflag = 0;                           /* Start with Log file closed */
SubsampleFlag = FALSE;
subsample = 1;

OpenHelp();


    do
    {
       comment = FALSE;
       if (flowthru)
       {
           flowthru = 0;
       }
       else if (BatchFlag)
       {
           if (abort_disp == 1)
               result = NULL;
           else
               result = fgets( CommandString, 127, BatchFile[BatchFileNumber]);
           if (result == NULL)
           {
               fclose (BatchFile[BatchFileNumber]);
               BatchFileNumber--;
               if (BatchFileNumber < 0)
                  BatchFlag = 0;
               if (SelectFlag)
               {
                  fclose (SelectFile);
                  SelectFlag = 0;
               }
               strcpy (CommandString, "DONE");
               abort_disp = 0;
           }

           for (i=0; i<strlen(CommandString); i++)
           {
              if (strncmp(&CommandString[i],";",1) == 0)
              {
                 if (i == 0)
                    comment = TRUE;
                 else
                    CommandString[i] = '\0';
              }
           }

           if (strnicmp (CommandString, "SEL", 3) == 0)
           {
               SelectFile = fopen (SelectName, "w");
               if (SelectFile == NULL)
               {
                  TextLine = 14;  TextSample = 1;
                  WriteText ("Error in opening Select file");
               }
               else
                  SelectFlag = 1;
           }
           if (strnicmp (CommandString, "PAU", 3) == 0)
           {
               GetKeywordInteger(CommandString, "PAU", 0, &i, &flag);
               if (flag < 1)
                   GetKeywordInteger(CommandString, "pau", 0, &i, &flag);

               if (flag==1)
               {
                   Delay( 1000*i );
               }
               else
               {
                   StatusLine(2,"Pausing - Type any key to continue...");
                   while(!kbhit); getch();
               }
           }

           /* write the batch command line */
           if (CommandString[strlen(CommandString)-1] == 0x0a)
               CommandString[strlen(CommandString)-1] =  '\0';

           if (PromptFlag)
              StatusLine(2,CommandString);
           if (DelayFlag)
              Delay(500);
       }
       else if (OneScreen)
       {
           abort_disp = 0;
           StatusLine(2,"Command: ");
           CommandFlag = TRUE;
           AcceptText (CommandString);
           CommandFlag = FALSE;
           strcpy (temp, "Command: ");
           strcat (temp, CommandString);
           strcat (temp, " ");

           LengthText (temp, TextHeight, &len);
           EraseText (TextLine, TextSample, TextHeight, len, 0);

       }
       else
       {
           abort_disp = 0;
           printf ("%s", "Command: ");
           gets (CommandString);
       }
       if (!comment)
       {
          strupr (CommandString);
          command = GetCommand (CommandString,
          "FIL DIS ERA SET HIS PAL STR CUR PRO REF HEL EXI QUI DON BAT SAV "
          "CD  DIR TYP SYS >   MEN LAB TEX OVE BRO PLO SER PRI NEG BRI DAR "
          "ENH SLA LOG PER COP MER MAS WIN ROT");
/* if you add more than 40 commands you must fix NCOMMANDS in IMDEF.H */
          switch (command)
          {   case  1 : GetFile();            break;
              case  2 : DisplayImage();       break;
              case  3 : ClearDisplay(0);      break;
              case  4 : SetValues();          break;
              case  5 : DoHistogram();        break;
              case  6 : DoPalette();          break;
              case  7 : DoStretch();          break;
              case  8 : DoCursor();           break;
              case  9 : DoProfile();          break;
              case 10 : DoRefresh();          break;
/*              case 11 : GiveHelp();           break; */
              case 11 : Help();           break;
              case 12 : /* exit command */    break;
              case 13 : /* quit command */    break;
              case 14 : /* done/null command*/break;
              case 15 : DoBatch();            break;
              case 16 : DoSave();             break;
              case 17 :

                  GetKeywordString (CommandString,"CD",".",temp, &flag);
                  result      = temp; /* change to char pointer */
                  segregs.ds  = FP_SEG(result);
                  inregs.h.ah = 0x3B;
                  inregs.x.dx = (unsigned)temp;
                  intdosx(&inregs,&outregs,&segregs);
                  break;

              case 18 :
              case 19 :
              case 20 :
              case 21 :

                  ReadPalette(CT);
                  DisplayOff();
                  save_refresh = RefreshLines;
                  FreeRefresh();
                  if (OpenFileFlag)
                  {
                      CloseImage (0, status);
                      if (BadStatus (status)) return(0);
                  }

                  comspec = getenv("COMSPEC");
                  if ((imbrowse = getenv("IMBROWSE")) != NULL)
                  {
                      strcpy(SwapName, imbrowse);
                      strcat(SwapName, "IMDISWAP.FIL");
                  }
                  if (command == 18 || command == 19)
                  {
                      strcpy(execute, "/C ");
                      strcat(execute,CommandString);
                  }
                  else
                  {
                      printf("Type \'exit\' to return to IMDISP.\n");
                      strcpy(execute,"");
                  }
             swap_return = swap(comspec, execute, &exec_return, SwapName);
                  if (swap_return == SWAP_NO_SAVE)
                  {
                  /*  We didn't have enough disk space to swap so clean up
                      the leftover swap file and try just spawning to DOS
                  */
                      remove( SwapName ) ;
                      printf( "Not much memory to work with...\n");
                      sys_return = system( execute );
                      if (sys_return != -1)
                          swap_return = SWAP_OK;

                  }
                  if (command == 18 || command == 19)
                  {
                      printf("Type any key to continue:");
                      while(!kbhit); getch();
                  }
                  DisplayOn();
                  WritePalette (CT);
                  if (OpenFileFlag)
                  {
                     OpenFileFlag = 0;
                     strcpy(CommandString,"FIL ");
                     strcat(CommandString,ImageFileName);
                     if (!ImageMemory)
                        strcat(CommandString," NOM");
                     GetFile();
                  }
                  if (save_refresh > 0)
                  {
                      if (RefreshLocation == VIRTUAL_FILE)
                      {
                         RefreshLines = save_refresh;
                         refresh = open( RefreshName, O_RDWR|O_BINARY,
                                         S_IREAD|S_IWRITE );
                      }
                      DoRefresh();
                  }
                  switch (swap_return)
                  {
                      case SWAP_OK:
                          StatusLine(0,"Successful, executed program returned.");
                          break;

                      case SWAP_NO_SHRINK:
                          StatusLine(0,"Unable to shrink DOS memory block.");
                          break;

                      case SWAP_NO_SAVE:
                          StatusLine(0,"Unable to save program to memory or disk.");
                          break;

                      case SWAP_NO_EXEC:
                          StatusLine(0,"DOS EXEC call failed.  Error is: ");
                          switch (exec_return)
                          {
                              case BAD_FUNC:
                                  StatusLine (1,"Bad function.");
                                  break;
                              case FILE_NOT_FOUND:
                                  StatusLine (1,"Program file not found.");
                                  break;
                              case ACCESS_DENIED:
                                  StatusLine (1,"Access to program file denied.");
                                  break;
                              case NO_MEMORY:
                                  StatusLine (1,"Insufficient memory to run program.");
                                  break;
                              case BAD_ENVIRON:
                                  StatusLine (1,"Bad environment.");
                                  break;
                              case BAD_FORMAT:
                                  StatusLine (1,"Bad format.");
                                  break;
                              default:
                                  StatusLine (1,"Unexpected error code.");
                                  break;

                          }
                  }
                  break;

              case 22: DoMenu();          break;
              case 23: DoLabel();         break;
              case 24: DoPosText();       break;
              case 25: DoOverlay();       break;
              case 26: DoBrowse();        break;
              case 27: DoPlot();          break;
              case 28: DoSerial();        break;
              case 29: DoPrint();         break;
              case 30: DoNegative();      break;
              case 31: Dark_Filter(0);    break; /* Added 05/91 - R. Baalke */
              case 32: Dark_Filter(1);    break; /* Added 05/91 - R. Baalke */
              case 33: DoFilter();        break; /* Added 05/91 - R. Baalke */
              case 34: DoSlant();         break; /* Added 05/91 - R. Baalke */
              case 35: DoLog();           break; /* Added 06/91 - A. Warnock */
              case 36: DoPerspect();      break; /* Added 06/91 - A. Warnock */
              case 37: DoCopy();          break; /* Added 06/91 - R. Baalke */
              case 38: DoMerge();         break; /* Added 06/91 - R. Baalke */
              case 39: DoMask();          break; /* Added 07/91 - R. Baalke */
              case 40: DoWindow();        break; /* Added 09/91 - R. Baalke */
              case 41: DoRotate();        break; /* Added 09/91 - R. Baalke */

              default : /* otherwise sound beep   */
              {
              /* check for drive change command */
                  if (strlen(CommandString) == 2 && CommandString[1] == ':')
                      bdos(14,(int)CommandString[0]-'A',0);
                  else if ((strncmp(CommandString,"PAU",3) != 0) &&
                           (strncmp(CommandString,"SEL",3) != 0))
                  {
                      StatusLine(2,"INVALID! ");
                      printf("\a");
                      for (i=0;i<5000;i++);
                  }
              }
          }
          if (logflag)
          {
              fputs( CommandString, LogFile );
              fputs( "\n", LogFile );
          }
       }
    } while ((command != 12) && (command != 13));
    GetKeywordSubcommand  (CommandString, "NOC", &clearflag);

/*  Deallocate any buffers */
    for (i=0; i<MAXBUFFERS; i++)
    {
       if (Buffers[i].location != NOWHERE)
          DeallocBuffer(i);
    }

    free (buffer1);
    free (buffer2);
    free (buffer3);
    free (buffer4);
    free (buffer5);
    free (buffer6);
    free (buffer7);
    free (outbuffer);

/*  Close up any open files */

    if (OpenFileFlag)
    {
       CloseImage (0, status);
       BadStatus (status);
    }

    if (logflag)
    {
       fflush( LogFile );
       fclose( LogFile );
    }

    for (i=0; i<BatchFileNumber+1; i++)
       fclose(BatchFile[i]);

/* Deallocate refresh buffer */

    FreeRefresh();
    DeallocRefresh();

/*  Restore the starting directory and drive */
    chdir( DefaultDir );
#ifdef __TURBOC__
    NumDrives = setdisk( DefaultDrive );
#else
    _dos_setdrive( DefaultDrive, &NumDrives );
#endif

/*  Back to text mode */
    if (clearflag != 1) DisplayOff();
    if (Mem_Used() != 0)
    {
       printf( "*** Memory list not empty ***\n" );
       Mem_Display( stdout );
       exit(1);
    }
    exit(0);
}

int DoRefresh(void)
/*  DoRefresh performs the refresh command.
    The whole display screen, or as much as there is in memory,
    is redisplayed.
*/
{
    int  line;
    unsigned char   *buffer;

    if (RefreshLines == 0)
    {
       StatusLine(0, "Refresh buffer inactive.");
       return(0);
    }
    if ((buffer = malloc(dispns)) == NULL)
    {
        FatalError( "Not enough memory for refr cmd.\n");
    }
    for (line = 1;  line <= RefreshLines;  line++)
    {
       GetRefresh (buffer, line, 1, dispns);
       DisplayLine (buffer, line, 1, dispns);
    }
    free (buffer);
}


int SetValues(void)
/*  SetValues performs the set command.
    The DN range parameters are gotten and printed out.
*/
{
    int    newns;
    int    swapflag, loflag, hiflag, refflag, dispflag, nsflag, broflag;
    char   dispstr[128];
    char   refresh_str[20];
    int    save_numDN;
    int    flag;

    GetKeywordString (CommandString, "SET", "",dispstr, &flag);
    GetKeywordLong (CommandString, "LO", DNlow , &DNlow, &loflag);
    if (loflag < 0)
       GetKeywordLong (CommandString, "DNL", DNlow , &DNlow, &loflag);
    GetKeywordLong (CommandString, "HI", DNhigh , &DNhigh, &hiflag);
    if (hiflag < 0)
       GetKeywordLong (CommandString, "DNH", DNhigh , &DNhigh, &hiflag);
    if (flag == 0)
     {
      sprintf (dispstr, "DN low : %5ld   DN high : %5ld             ", DNlow, DNhigh);
      StatusLine(1,dispstr);
      return(0);
     }

    GetKeywordString (CommandString, "REF", "", refresh_str, &refflag);
    if (refflag > 0)
    {
       if (strcmp( refresh_str, "ON") == 0)
       {
           RefreshLines = AllocRefresh(0);
           itoa(RefreshLines, refresh_str, 10);
           strcpy( dispstr, refresh_str);
           strcat( dispstr, " lines allocated in refresh buffer");
           StatusLine (1, dispstr);
       }
       else
       {
           FreeRefresh();
           DeallocRefresh();
           RefreshLines = 0;
       }
    }

    /* Added 07/20/91 - Ron Baalke (Viking 1 anniversary!) */
    GetKeywordString (CommandString, "PRO", "", dispstr, &flag);
    if (flag > 0)
    {
       if (strcmp(dispstr, "ON") == 0)
          PromptFlag = TRUE;
       else
          PromptFlag = FALSE;
    }

    /* Added 07/20/91 - Ron Baalke (Viking 1 anniversary!) */
    GetKeywordString (CommandString, "DEL", "", dispstr, &flag);
    if (flag > 0)
    {
       if (strcmp(dispstr, "ON") == 0)
          DelayFlag = TRUE;
       else
          DelayFlag = FALSE;
    }

    GetKeywordInteger (CommandString, "NS", IMCB[0].ns , &newns, &nsflag);
    if (nsflag > 0)
    {
        IMCB[0].ns = newns;
        IMCB[0].reclen = newns;
    }

    GetKeywordSubcommand  (CommandString, "SWA", &swapflag);
    if (swapflag > 0) ByteSwap = TRUE;
    GetKeywordSubcommand  (CommandString, "UNSWA", &swapflag);
    if (swapflag > 0) ByteSwap = FALSE;

    GetKeywordString  (CommandString, "BRO", "C:\\BROWSE.CMD" , BrowseName, &broflag);
    GetKeywordString  (CommandString, "SEL", "C:\\BROWSE.SEL" , SelectName, &broflag);
    GetKeywordString  (CommandString, "PAL", ".\\PALETTES"    , PalettePath,&broflag);
    GetKeywordString  (CommandString, "BUF", "C:\\"           , BufferFileName,&broflag);

    GetKeywordString  (CommandString, "DIS", "vga", dispstr, &dispflag);
    if (dispflag > 0)
    {
       ReadPalette(CT);              /* Save current palette */
       save_numDN = numDN;
       DisplayOff();
       if (stricmp(dispstr, "cga"      ) == 0)
           DisplayDevice = CGA200;
       else if (stricmp(dispstr, "ega480"   ) == 0)
           DisplayDevice = EGA480;
       else if (stricmp(dispstr, "ega"      ) == 0)
           DisplayDevice = EGA350;
       else if (stricmp(dispstr, "pga"      ) == 0)
           DisplayDevice = PGA;
       else if (stricmp(dispstr, "vga320"   ) == 0)
           DisplayDevice = VGA200;
       else if (stricmp(dispstr, "vga"      ) == 0)
           DisplayDevice = VGA480;
       else if (stricmp(dispstr, "orchid1024") == 0)
           DisplayDevice = ORCHID768;
       else if (stricmp(dispstr, "orchid800") == 0)
           DisplayDevice = ORCHID600;
       else if (stricmp(dispstr, "orchid"   ) == 0)
           DisplayDevice = ORCHID480;
       else if (stricmp(dispstr, "evga640"  ) == 0)
           DisplayDevice = EVGA640;
       else if (stricmp(dispstr, "evga512"  ) == 0)
           DisplayDevice = EVGA512;
       else if (stricmp(dispstr, "evga800"  ) == 0)
           DisplayDevice = EVGA800;
       else if (stricmp(dispstr, "ati640"   ) == 0)
           DisplayDevice = ATI640;
       else if (stricmp(dispstr, "ati800"   ) == 0)
           DisplayDevice = ATI800;
       else if (stricmp(dispstr, "ati1024"  ) == 0)
           DisplayDevice = ATI1024;
       else if (stricmp(dispstr, "paradise" ) == 0)
           DisplayDevice = PARADISE;
       else if (stricmp(dispstr, "trident" ) == 0)
           DisplayDevice = TRIDENT;
       else if (stricmp(dispstr, "vesa800"  ) == 0)
           DisplayDevice = VESA800;
       else if (stricmp(dispstr, "vesa640"  ) == 0)
           DisplayDevice = VESA640;
       else if (stricmp(dispstr, "vesa"     ) == 0)
           DisplayDevice = VESA;
       else if (stricmp(dispstr, "ev629"    ) == 0)
           DisplayDevice = EV629;
       else if (stricmp(dispstr, "test"     ) == 0)
           DisplayDevice = TEST;
#ifdef __TURBOC__
       else if (stricmp(dispstr, "bios"     ) == 0)
           DisplayDevice = VGA480;
#else
       else if (stricmp(dispstr, "bios"     ) == 0)
           DisplayDevice = BIOS;
#endif

       DisplayOn();
       if (save_numDN == numDN)   /* if color count match */
          WritePalette (CT);      /* restore palette */
       else
          WritePalette(DefaultPalette);
       if (RefreshLines > 0)
       {
          EraseRefresh();
       }
    }

    if (DNlow > 32767)  DNlow = 32767;
    if (DNhigh < DNlow+2)  DNhigh = DNlow + 2;

    if (loflag > 0 ||  hiflag > 0)
    {
       sprintf (dispstr, "DN low : %5ld   DN high : %5ld                 ", DNlow, DNhigh);
       StatusLine(1,dispstr);
    }
}


int DoHistogram(void)
/*  DoHistogram performs the histogram command.
    First the parameter are scanned.  Then the image is read
    in and the histogram calculated.  Finally DisplayHistogram
    is called to plot it on the screen.  Different pieces of code
    are used for different pixel formats.
*/
{
    int            hnld, hnsd, hsl, hss;
    int            hnldd, hnsdd;
    int            hsubsample;
    int            line, samp, i, j;
    int            DN, scale, flag, VoyFlag, VikFlag;
    long int       histbuf[256];
    char           status[128], dispstr[80], ch;
    unsigned char  *buffer, *tempbuf;
    int            *intbuf;
    int            nl_save, ns_save;
    int            buffer_flag = 0;
    int            index;

    if (OpenFileFlag == 0) 
    {
       StatusLine(0,"Use the \'FILE fname\' command to select image first");
       return(0);
    }

    if (bitsperpix == 32)
    {
    /* added by AEE: 01/17/90 */
       StatusLine(0,"No histogram for 32 bits per pixel");
       return(0);
    }

    DrawBox (dispnl-45, 1, 46, dispns, 0);
    for (DN = 0;  DN <= 255;  DN++)
       histbuf[DN] = 0;

/*  Check to see if buffer is to be displayed - Ron Baalke - 07/25/91 */
    GetKeywordString (CommandString, "HIS", " ",dispstr, &flag);

    if ((flag)                 &&
        (strlen(dispstr) == 1) &&
        (dispstr[0] >= 'A')    &&
        (dispstr[0] <= 'Z'))
    {
       buffer_flag = 1;
       index = dispstr[0] - 'A';
       if (Buffers[index].location == NOWHERE)
       {
          strcpy(status,"A");
          status[0] += index;
          sprintf(dispstr,"Buffer %s is empty",status);
          StatusLine(0,dispstr);
          return(1);
       }
       nl_save = nl;
       ns_save = ns;
       nl = Buffers[index].dispnl;
       ns = Buffers[index].dispns;
    }

    GetKeywordSubcommand  (CommandString, "COL", &flag);
    if (flag >= 0) Color_Hist= 1;
      else Color_Hist = 0;

    GetKeywordSubcommand  (CommandString, "CEN", &flag);
    if (flag == 0) 
    {
       hsl = lastsl;
       hss = lastss;
       hnld = nld;
       hnsd = nsd;
       if (hnld+hsl-1 > nl)
           hnld = nl-hsl+1;
       if (hnsd+hss-1 > ns)
           hnsd = ns-hss+1;
       hnldd = zoom*hnld / subsample;
       hnsdd = zoom*hnsd / subsample;
       if (hnldd+sldd-1 > dispnl)
           hnldd = dispnl-sldd+1;
       if (hnsdd+ssdd-1 > dispns)
           hnsdd = dispns-ssdd+1;
       hnld = subsample*hnldd / zoom;
       hnsd = subsample*hnsdd / zoom;
       hnldd = hnldd/zoom;
       hnsdd = hnsdd/zoom;
       hsubsample = subsample;
    }
    else 
    {
        GetKeywordInteger (CommandString, "NL", nl, &hnld, &flag);
        GetKeywordInteger (CommandString, "NS", ns, &hnsd, &flag);
        GetKeywordInteger (CommandString, "SL", 1, &hsl, &flag);
        GetKeywordInteger (CommandString, "SS", 1, &hss, &flag);
        GetKeywordInteger (CommandString, "SUB", 1, &hsubsample, &flag);
        GetKeywordString  (CommandString, "VOY", " ",status,&VoyFlag);
        GetKeywordString  (CommandString, "VIK", " ",status,&VikFlag);

        if (hnld+hsl-1 > nl)
            hnld = nl-hsl+1;
        if (hnsd+hss-1 > ns)
            hnsd = ns-hss+1;
        hnldd = hnld / hsubsample;
        hnsdd = hnsd / hsubsample;
    }

    strcpy(status,"");
    if ((bitsperpix == 16) || (bitsperpix == 32))      /*  16 bit images */
    {
        if ((intbuf = (int *)malloc(2*hnsd)) == NULL)
        {
            FatalError( "Not enough memory for intbuf 16.\n");
        }
       scale = max( 1, ((DNhigh - DNlow) / 255) );
       line = hsl;
       for (i = 0;  i < hnldd;  i++)
       {
           if (ImageLocation == VIRTUAL_FILE)
              ReadLine (0, intbuf, line, hss, hnsd, status);
           else
              ReadXMSLine (intbuf, line, hss, hnsd);
           if (BadStatus(status))
           {
               free(intbuf);
               return(1);
           }
           samp = 0;
           for (j = 0;  j < hnsdd;  j++)
           {
               DN = intbuf[samp];
               if (DN >= DNhigh)
                   DN = DNhigh;
               else if (DN <= DNlow)
                   DN = DNlow;
              DN = ((long)DN-DNlow) / scale;
               DN = min( DN, 255);
               DN = max( DN, 0);
               (histbuf[DN])++;
               samp += hsubsample;
           }
           line += hsubsample;
/*
   keyboard abort added
*/
               if (kbhit())  /* abort disp if keypressed mdm 2/19/88*/
               {
                   if ((ch = getch()) == 0) ch = 0x80 | getch();
                   if (ch != 17) /* don't abort if cntl q */
                   {
                       j=hnldd;
                       free(intbuf);
                       return(1);
                   }
               }
       }
       if (abort_disp != 1)
           DisplayHistogram (histbuf, 256, DNlow, DNlow+256*scale);

       free(intbuf);
    }

    else if (bitsperpix == 8)          /*  8 bit images */
    {

       if ((buffer = malloc(hnsd)) == NULL)
        {
            FatalError( "Not enough memory for intbuf 8.\n");
        }
       if ((hss==1 && hnsd==800 && hsl==1 && hnld == 800 && hsubsample == 1
             && VoyFlag == 0) ||
           (hss==1 && hnsd==1204 && hsl==1 && hnld == 1056 && hsubsample == 1
             && VikFlag == 0) ||
           (hss==1 && hsl == 1 && hnld == nl && hnsd == ns && hsubsample == 1
             && Histogram > 0L ))
          /* read  histogram from file*/
       {
           /* location of vgr histo */
           if (VoyFlag == 0 && Histogram == 0)
               Histogram = 671496L + (Microsoft*2048);

           /* location of vik histo */
           if (VikFlag == 0 && Histogram == 0)
               Histogram = 1337472L + (Microsoft*2048);

           lseek(FCB[0].handle,Histogram,SEEK_SET);
           read(FCB[0].handle,(char *)histbuf,1024);
       }
       else
       {
           line = hsl;
           for (i = 0;  i < hnldd;  i++)
           {
               if (buffer_flag)
                  GetBuffer(index, buffer, line, hss, hnsd);
               else if (ImageLocation == VIRTUAL_FILE)
                  ReadLine (0, buffer, line, hss, hnsd, status);
               else
                  ReadXMSLine (buffer, line, hss, hnsd);
               if (BadStatus(status))
               {
                   free (buffer);
                   return(1);
               }
               samp = 0;
               for (j = 0;  j < hnsdd; j++)
               {
                   (histbuf[buffer[samp]])++;
                   samp += hsubsample;
               }
               line += hsubsample;
/*
   keyboard abort added
*/
               if (kbhit())  /* abort disp if keypressed mdm 2/19/88*/
               {
                   if ((ch = getch()) == 0) ch = 0x80 | getch();
                   if (ch != 17) /* don't abort if cntl q */
                   {
                       j=hnldd;
                       free (buffer);
                       return(1);
                   }
               }
           }
       }
       if (abort_disp != 1)
           DisplayHistogram (histbuf, 256, 0, 255);

       free (buffer);
    }

    else                           /*  4 bit and 1 bit images */
    {

       if ((buffer = malloc(hnsd)) == NULL)
        {
            FatalError( "Not enough memory for intbuf 4.\n");
        }
       if ((tempbuf = malloc(hnsd)) == NULL)
        {
            FatalError( "Not enough memory for temp buf.\n");
        }
       line = hsl;
       for (i = 0;  i < hnldd;  i++)
       {
           if (ImageLocation == VIRTUAL_FILE)
              ReadLine (0, tempbuf, line, hss, hnsd, status);
           else
              ReadXMSLine (tempbuf, line, hss, hnsd);
           if (BadStatus(status))
           {
               free (buffer);
               free (tempbuf);
               return(1);
           }
           ConvertLine (tempbuf, buffer, bitsperpix, 8, hnsd, status);
           samp = 0;
           for (j = 0;  j < hnsdd; j++)
           {
               (histbuf[buffer[samp]])++;
               samp += hsubsample;
           }
           line += hsubsample;
/*
   keyboard abort added
*/
               if (kbhit())  /* abort disp if keypressed mdm 7/22/91*/
               {
                   if ((ch = getch()) == 0) ch = 0x80 | getch();
                   if (ch != 17) /* don't abort if cntl q */
                   {
                       j=hnldd;
                       free(buffer);
                       free(tempbuf);
                       return(1);
                   }
               }
        }
        DisplayHistogram (histbuf, 16, 0, 15);

        free (buffer);
        free (tempbuf);
    }

    if (buffer_flag)
    {
       nl = nl_save;
       ns = ns_save;
    }

}


int DoPalette(void)
/***   DoPalette performs the palette command.

***/
{
    int    pseudo, flag, numdiddle, diddleflag;
    int    i, color, saveflag, loadflag, showflag, eraseflag;
    int    cycleflag, cycle;
    char   savename[64], loadname[64];
    FILE   *unit;
    char   tempdir[67];                    /* starting path */
    unsigned tempdrive, tempdrives;         /* starting drive */
    int    wildcard, sel;

    GetKeywordInteger (CommandString, "PS", 0, &pseudo, &flag);
    GetKeywordInteger (CommandString, "EDI", numDN, &numdiddle, &diddleflag);
    if (diddleflag < 0)
       GetKeywordInteger (CommandString, "DID", numDN, &numdiddle, &diddleflag);
    GetKeywordString (CommandString, "SAV", "IMDISP.PAL", savename, &saveflag);
    GetKeywordString (CommandString, "LOA", "", loadname, &loadflag);
    GetKeywordSubcommand (CommandString, "DIS", &showflag);
    GetKeywordSubcommand (CommandString, "ERA", &eraseflag);
    GetKeywordInteger (CommandString, "CYC", 0, &cycle, &cycleflag);

    if (saveflag >= 0)
    {
       if ( (unit = fopen (savename, "w")) == NULL )
           return(0);
       ReadPalette (CT);
       fprintf (unit, "CCSD3ZF0000100000001NJPL3IF0PDS200000001 = SFDU_LABEL\012");
       fprintf (unit, "RECORD_TYPE                     = STREAM\012");
       fprintf (unit, "FILE_RECORDS                    = %3d\012", 13+numDN);
       fprintf (unit, "OBJECT                          = TABLE\012");
       fprintf (unit, " TABLE_ROWS                      = %3d\012", numDN);
       fprintf (unit, " ROW_COLUMNS                     = 4\012");
       fprintf (unit, " COLUMN_NAME                     = (COLOR_NUMBER,\012");
       fprintf (unit, "                                    RED_VALUE,\012");
       fprintf (unit, "                                    GREEN_VALUE,\012");
       fprintf (unit, "                                    BLUE_VALUE)\012");
       fprintf (unit, " COLUMN_TYPE                     = (INTEGER,INTEGER,INTEGER,INTEGER)\012");
       fprintf (unit, "END_OBJECT\012");
       fprintf (unit, "END\012");

       for (i = 0;  i < numDN;  i++)
           fprintf (unit, "%4d,%4d,%4d,%4d\012", i, CT[i].r, CT[i].g, CT[i].b);
       fclose (unit);
    }

    else if (loadflag >=0)      /* Ron Baalke - 04/07/91 */
    {
       wildcard = FALSE;
       if (loadflag != 0)
       {
          for (i=0; i<strlen(loadname); i++)
              if (strncmp(&loadname[i],"*",1) == 0)
              {
                 wildcard = TRUE;
                 strcpy(palMask,loadname);
              }
       }
     getcwd( tempdir, 66);                   /* Save starting directory*/
#ifdef __TURBOC__
     tempdrive = getdisk() + 'A';
#else
     _dos_getdrive( &tempdrive );
#endif
     if ((loadflag > 0) && (ChangeDir(loadname) == 0))
        strcpy(PalettePath,loadname);

     if ((loadflag == 0) || wildcard)
          GetPalette();
       else
          LoadPalette(loadname);

     chdir( tempdir );                       /* Move back to directory */
#ifdef __TURBOC__
     tempdrives = setdisk( tempdrive );
#else
     _dos_setdrive( tempdrive, &tempdrives );
#endif
    }

    else if (diddleflag >= 0)
    {
       DiddlePalette (numdiddle);
    }
    else if (showflag >= 0)
    {
       ShowPalette (numdiddle);
    }
    else if (eraseflag >= 0)
    {
       ErasePalette (numdiddle);
    }
    else if (cycleflag >= 0)  /* Added palette cycling - Ron Baalke - 2/4/91 */
    {
       ReadPalette (CT);
       if (!CyclePalette (cycle))
          WritePalette (CT);
    }


    else if (flag < 0) /* pal command with no args loads default */
       WritePalette(DefaultPalette);
    else
    {
       switch (pseudo)
       {
       case 0 :

               for (i = 0;  i < numDN;  i++)
               {
                   color = (16*i)/numDN;
                   if (color > 15)  color = 15;
                   CT[i].r = PScolor[color][0];
                   CT[i].g = PScolor[color][1];
                   CT[i].b = PScolor[color][2];
               }
               WritePalette (CT);
               break;
 
       case 1 :
 
               if (DisplayDevice == ATI1024)
               {
                  for (i = 0; i < 256; i++)
                  {
                     CT[i].r = i;
                     CT[i].g = i;
                     CT[i].b = i;
                  }
                  WritePalette (CT);
               }
               else
                  Stretch (0, numDN-1);
               break;
 
       case 2 :
 
               for (i = 0;  i < numDN;  i++)
               {
                   color = (16*i)/numDN;
                   if (color > 15)  color = 15;
                   CT[i].r = Scolor[color][0];
                   CT[i].g = Scolor[color][1];
                   CT[i].b = Scolor[color][2];
               }
               WritePalette (CT);
               break;
 
       case 3 :
 
               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = Redorange[i][0];
                   CT[i].g = Redorange[i][1];
                   CT[i].b = Redorange[i][2];
               }

              if (numDN == 16)
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }

               WritePalette (CT);
               break;
 
       case 4 :
 
               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = Prism[i][0];
                   CT[i].g = Prism[i][1];
                   CT[i].b = Prism[i][2];
               }

              if (numDN == 16)
               {
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }
               /* special fix to display text */
               CT[15].r=255; CT[15].g=255; CT[15].b=255;
               }
               WritePalette (CT);
               break;

       case 5 :

               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = Topo[i][0];
                   CT[i].g = Topo[i][1];
                   CT[i].b = Topo[i][2];
               }

              if (numDN == 16)
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }

               WritePalette (CT);
               break;

       case 6 :

               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = Contour[i][0];
                   CT[i].g = Contour[i][1];
                   CT[i].b = Contour[i][2];
               }

              if (numDN == 16)
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }

               WritePalette (CT);
               break;

       case 7 :

               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = BGRY[i][0];
                   CT[i].g = BGRY[i][1];
                   CT[i].b = BGRY[i][2];
               }

              if (numDN == 16)
               {
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }
               /* special fix to display text */
               CT[15].r=255; CT[15].g=255; CT[15].b=255;
               }

               WritePalette (CT);
               break;
       case 8 :

               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = Glasses1[i][0];
                   CT[i].g = Glasses1[i][1];
                   CT[i].b = Glasses1[i][2];
               }

              if (numDN == 16)
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }

               WritePalette (CT);
               break;
       case 9 :

               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = Glasses2[i][0];
                   CT[i].g = Glasses2[i][1];
                   CT[i].b = Glasses2[i][2];
               }

              if (numDN == 16)
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }

               WritePalette (CT);
               break;

       case 10:

               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = Perspect[i][0];
                   CT[i].g = Perspect[i][1];
                   CT[i].b = Perspect[i][2];
               }

              if (numDN == 16)
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }

               WritePalette (CT);
               break;

       case 11:

               for (i = 0;  i < 256;  i++)
               {
                   CT[i].r = Hubble[i][0];
                   CT[i].g = Hubble[i][1];
                   CT[i].b = Hubble[i][2];
               }

              if (numDN == 16)
                for (i=0;i<16;i++)
                 {
                  CT[i].r = CT[i*16+8].r;
                  CT[i].g = CT[i*16+8].g;
                  CT[i].b = CT[i*16+8].b;
                 }

               WritePalette (CT);
               break;

       }
    }
}




int DoStretch(void)
/*  DoStretch performs the stretch command.
*/
{
    int DNlow, DNhigh, flag;

    GetKeywordInteger (CommandString, "LO", 0, &DNlow, &flag);
    if (flag < 0)
       GetKeywordInteger (CommandString, "DNL", 0 , &DNlow, &flag);
    GetKeywordInteger (CommandString, "HI", numDN-1, &DNhigh, &flag);
    if (flag < 0)
       GetKeywordInteger (CommandString, "DNH", numDN-1 , &DNhigh, &flag);
    Stretch (DNlow, DNhigh);
}

int DoCursor(void)
/*  DoCursor performs the cursor command.
    After cursor mode the image line and sample are calculated and displayed.
*/
{
    int    line, samp, i, len, DN;
    char   dispstr[64];
    int    line1, samp1;
    int    flag1, flag2;

    GetKeywordInteger (CommandString, "DSL", 1, &line1, &flag1);
    GetKeywordInteger (CommandString, "DSS", 1, &samp1, &flag2);
/*    if ((flag1 != -1) && (flag2 = -1)) */
    if ((flag1 != -1) && (flag2 != -1))
    {
       line = line1;
       samp = samp1;
    }
    else
    {
       StatusLine(1,"Mouse, arrow keys move; click mouse or <CR> to exit.");
       DN = MoveCursor(&line,&samp);
       sprintf (CommandString, "CURSOR DSL %d DSS %d DN %d",line,samp,DN);
    }

    centerline = subsample*(line-sldd) / zoom + sl;
    centersamp = subsample*(samp-ssdd) / zoom + ss;

    if (PromptFlag)
    {
       sprintf (dispstr, "Line : %5d   Sample : %5d   DN : %8d", centerline,
               centersamp, DN);
       StatusLine(1,dispstr);
    }
    return(0);
}

int DoProfile(void)
/*  DoProfile performs the profile command.
*/
{
    DrawBox (dispnl-45, 1, 46, dispns, 0);
    Profile();
}

int DoSave(void)
/* should ask for nl,ns if there is no open image file */

{
/* DoSave saves the current display to a disk file */

    char buffer[MAXDISPNS];
    char  filename[128], status[128], dispstr[128];
    unsigned char *linebuf, *outbuf;
    int   i, j, k, flag;
    int   ssl, sss, snl, sns, sbitsperpix;
    int   DN;
    int   defnl, defns;
    int   dispbits = 1;
    long   npix_to_file;

    GetKeywordSubcommand (CommandString, "REFRESH", &flag);
    if ((flag != -1) && (RefreshLines > 0))
    {
       Screen2Refresh();
       return(0);
    }

    if (OpenFileFlag == 0)
    {
       StatusLine(0,"Use the \'FILE fname\' command to select image");
       return(0);
    }

    if (numDN ==  16) dispbits = 4;
    if (numDN == 256) dispbits = 8;

    defnl = (dispnl > nl) ? nl : dispnl; /* default nl and ns */
    defns = (dispns > ns) ? ns : dispns;

    GetKeywordInteger (CommandString, "SL", 1, &ssl, &flag);
    GetKeywordInteger (CommandString, "SS", 1, &sss, &flag);
    GetKeywordInteger (CommandString, "NL", defnl-ssl+1, &snl, &flag);
    GetKeywordInteger (CommandString, "NS", defns-sss+1, &sns, &flag);
    GetKeywordInteger (CommandString, "BP", dispbits, &sbitsperpix, &flag);
    GetKeywordString  (CommandString, "SAV", " ", filename, &flag);

    GetKeywordString  (CommandString, "FOR", "", dispstr, &flag);

    if (strcmp(dispstr,"GIF") == 0)
       ImageFormat = GIF_FORMAT;
    else if (strcmp(dispstr,"RAW") == 0)
       ImageFormat = RAW_FORMAT;
    else
       ImageFormat = PDS_FORMAT;

  /* make sure parameters don't get interpreted as a filemask. */
    if (strcmp(filename,"SL") == 0 || strcmp(filename,"SS") == 0 ||
        strcmp(filename,"NL") == 0 || strcmp(filename,"NS") == 0 ||
        strcmp(filename,"BP") == 0)
    {
       BadStatus ("Must give name of image file, (eg. SAVE image.img)");
       return(0);
    }

    /* Refresh the bottom of the screen to cover up the command prompt */
    /* Ron Baalke - 06/15/91 */

    if (RefreshLines > 0)
    {
       for (i = dispnl-2*TextHeight-5; i<=dispnl; i++)
       {
         GetRefresh(buffer,i,1,dispns);
         DisplayLine(buffer,i,1,dispns);
       }
    }

    OpenImage (filename, 1, "write", &snl, &sns, &sbitsperpix, status);
    if (BadStatus (status))
       return(0);

    if ((linebuf = malloc(sns*2)) == NULL)
    {
        FatalError( "Not enough memory for linebuf save.\n");
    }
    if ((outbuf = malloc(sns*2)) == NULL)
    {
        FatalError( "Not enough memory for outbuf save.\n");
    }
    j = 1;
    npix_to_file = ((long) sns * sbitsperpix) / 8;
    for (i = ssl;i < ssl+snl; i++)
    {
    /* check here for going off screen */
       for (k=sss; k<sns+sss; k++)
       {
           ReadPixel( i, k, &DN);
           linebuf[ k-sss ] = DN;
       }
       if (sbitsperpix != 8)
       {
           ConvertLine( linebuf, outbuf, 8, sbitsperpix, sns, status);
           WriteLine( 1, outbuf, j, 1, (int)npix_to_file, status);
       }
       else
       {
            WriteLine( 1, linebuf, j, 1, sns, status);
       }
       if ((j % 25) == 0)
       {
           sprintf( dispstr, "Saved image line %d", j);
           StatusLine(0,dispstr);
       }
       j++;
    }
    free(linebuf);
    free(outbuf);

    CloseImage(1,status);
    sprintf( dispstr, "Image saved to file %s, %dx%d resolution",
             filename, ssl+snl-1, sns+sss-1);
    StatusLine(0,dispstr);
}

int DoBatch(void)
{
    char BatchName[32];
    int  flag = 0;
    char buffer[81];

    GetKeywordString (CommandString, "BAT", "BATCH.CMD", BatchName, &flag);
/*    ClearDisplay(0); */
    BatchFileNumber++;
    BatchFile[BatchFileNumber] = fopen(BatchName,"r");
    if (BatchFile[BatchFileNumber] == NULL)
    {
       BatchFileNumber--;
       StatusLine(0,"File not found.");
    }
    else
    {
    /* check for SFDU labels */
       BatchFlag = 1;
       fgets(buffer, 80, BatchFile[BatchFileNumber]);
       if (strnicmp(buffer,"CCSD",4) == 0 ||
           strnicmp(buffer,"NJPL",4) == 0)
           do
           {
               fgets(buffer,80,BatchFile[BatchFileNumber]);
               if (strnicmp(buffer,"END",3) == 0 &&
                  (strlen(buffer) <= 5 || buffer[3] == ' '))
                     break;
           } while (!feof(BatchFile[BatchFileNumber]));
       else
           fseek(BatchFile[BatchFileNumber], 0L, SEEK_SET);
    }
}


int DoLabel(void)
{
    FILE   *labels;
    int    i;
    char   buffer[81];
    int    line_count=0;
    int    ch;
    int    done;

    if (OpenFileFlag == 0)
    {
       StatusLine(0,"Use the \'FILE fname\' command to select image");
       return(0);
    }

    ReadPalette(CT);
    DisplayOff();
    for (i=0; i<24; i++)
       printf("\n");

    if (LabelFileName[0] != '\0')
       labels = fopen(LabelFileName,"r");
    else
    {
       lseek( FCB[0].handle, 0L, SEEK_SET);
       labels = fdopen(FCB[0].handle,"r");
    }

    done = FALSE;
    while (!done)
    {
       fgets(buffer,81,labels);
       printf("%s",buffer);
       line_count++;
       if (line_count == 22)
       {
          printf("\n<< Hit Q to quit, or any other key to continue... >>\n");
          ch = getch();
          if ((ch == 81) || (ch == 113)) done = TRUE;
          printf("\n");
          line_count=0;
       }
       if (feof(labels) || (strnicmp(buffer,"END",3) == 0 &&
          (strlen(buffer) <= 5 || buffer[3] == ' ')))
       {
           done = TRUE;
           printf( "\n<< Press any key to return to command line >>\n" );
           getch();
       }
    }
    if (LabelFileName[0] != '\0')
       fclose(labels);
 
    if (!done)
    {
      printf( "\n<< Press any key to return to command line >>\n" );
      getch();
    }
    DisplayOn();
    WritePalette(DefaultPalette);
}



int DoMenu(void)
{
    FILE   *MenuFile;
    char   temp[40], string[81], writebuf[81], choice[10], *fileopt[50];
    int    flag = 0, done = 0, index = 0, toplist = 1, lowlist = PAGELEN;
    int    charcnt=81, i, numchoice;

    GetKeywordString(CommandString, "MEN", "MENU.MNU",temp, &flag);
    if (flag == 1)
        strcpy(MenuName, temp);
    if ((MenuFile = fopen(MenuName,"r"))==NULL)
    {
       StatusLine(0,"File not found.");
       return(0);
    }
    fseek(MenuFile, 0L, SEEK_SET);
    do
    {
       fgets(string, charcnt, MenuFile);
       if (strnicmp(string,"CCSD",4) == 0 ||
           strnicmp(string,"NJPL",4) == 0)
       {
           do
           {
               fgets(string,charcnt,MenuFile);
               if (strnicmp(string,"END",3) == 0 &&
                  (strlen(string) <= 5 || string[3] == ' ')) break;
           } while (!feof(MenuFile));
       fgets(string,charcnt,MenuFile);
       }
       for (i = 1; i < 81; i++)
           if (string[i] == '\n')
               string[i] = '\0';
       index += 1;
       if ((fileopt[index] = malloc(strlen(string)+1))==NULL)
        {
            FatalError( "Not enough memory for fileopt.\n");
        }
       strcpy(fileopt[index],string);
    } while (!feof(MenuFile));

    fclose(MenuFile);
    index--;  /* Take off the END from the menu list */

    while( !done )
    {
       ClearDisplay(0);
       TextLine = TextHeight + 5; TextSample = 10;
       WriteText("Menu Selection");
       if (index < PAGELEN)
           lowlist = index;

       for (i = toplist; i <= lowlist; i++)
       {
           sprintf( writebuf, "%d) %s", i, fileopt[i]);
           WriteText(writebuf);
       }
       WriteText("Choose: <#>, N(ext page), P(revious page), Q(uit)");
       TypeText("Option>> ");
       /*TextLine-=15;*/ TextSample = 90;
       AcceptText(choice);

       switch(choice[0])
       {

       case 'N':
       case 'n':

           toplist = ((toplist + PAGELEN) > index) ? index : toplist + PAGELEN;
           lowlist = ((toplist + PAGELEN) > index) ? index : toplist + PAGELEN;
           break;

       case 'P':
       case 'p':

           toplist = ((toplist - PAGELEN) < 1) ? 1 : toplist - PAGELEN;
           lowlist = ((toplist + PAGELEN) > index) ? index : toplist + PAGELEN-1;
           break;

       case 'Q':
       case 'q':

           ClearDisplay(0);
           done = 1;
           break;

       default:

           if (isdigit(choice[0]))
           {
               numchoice= atoi(choice);
               if (numchoice > 0 && numchoice <= index)
               {
                   if ((strnicmp(fileopt[numchoice]+
                        strlen(fileopt[numchoice])-4,".pal",4))== 0)
                           strcpy(CommandString,"PAL LOAD ");

                   else if ((strnicmp(fileopt[numchoice]+
                        strlen(fileopt[numchoice])-4,".cmd",4))== 0)
                           strcpy(CommandString,"BATCH ");
                   else
                       strcpy(CommandString,"FILE ");

                   strcat( CommandString, fileopt[numchoice]);
                   flowthru = 1;
                   done     = 1;
                   ClearDisplay(0);
               }
               else
                   WriteText("Invalid input!!!!");
           }
           break;
       }
    }
    for (i = 1; i <= index+1; i++)
       free(fileopt[i]);
}

int DoPosText(void)
{
    char   string[81], dispstr[81], *lquotptr, *fquotptr;
    int    flag = 0;
    int    line = 0, samp = 0;
    int    noprompt_flag;
    int    i;


    GetKeywordInteger (CommandString, "LIN", 0, &line, &flag);
    GetKeywordInteger (CommandString, "SAM", 0, &samp, &flag);
    GetKeywordSubcommand (CommandString, "NOP", &noprompt_flag);
    GetKeywordString(CommandString, "TEX", "", string, &flag);
    if (strncmp(string,"NOP",3) == 0) flag = 0;
    if (!flag)
    {
       StatusLine(2,"Please enter text:");
       AcceptText(string);
       strcpy(dispstr,"Please enter text:");
       strcat(dispstr,string);

       LengthText (dispstr, TextHeight, &line);
       EraseText (TextLine, TextSample, TextHeight, line, 0);

       if (noprompt_flag == -1)
          DoRefresh();
    }
    else
    {
       fquotptr = strpbrk(CommandString,"\'\"");
       lquotptr = strrchr(CommandString,'\'');
       if (lquotptr == NULL)
       {
          lquotptr = strrchr(CommandString,'\"');
          CommandString[strlen(CommandString)] = '\0';
       }
       else
           strcpy(lquotptr,"\0");
       strcpy(string,++fquotptr);
    }

    if (line == 0 || samp == 0)
    {
       if (noprompt_flag == -1)
          StatusLine(1,"Arrow keys move; +,- changes cursor increment; '.' to exit.");
       MoveCursor(&line,&samp);
       if (noprompt_flag == 1)
       {
          dispstr[0] = '\0';
          for (i=0; i<80; i++) strcat(dispstr," ");
          StatusLine(2,dispstr);
       }
    }

    TextLine = line - TextHeight;
    if (TextLine <= 0) TextLine = 1;
    TextSample = samp;
    IntoMem = 0;
    TypeText(string);
    TextSample = samp;

    /* If noprompt_flag is raised, then pause until a key is hit by the
       user.  This will delay the command line prompt from being delayed */

    if (noprompt_flag == 1)
       i = getch();

/*
    IntoMem = 1;
    TypeText(string);
    DoRefresh();
*/
}

int DoOverlay(void)
{
    int    dn, grids, i, j, x, y, flag, start, incr, defdn, offset = 0;
    char   label[20];
    float  delta, mlat, mlon, hold;

    if (zoom != 1 || subsample != 1)
    {
       StatusLine(0,"Can't overlay zoomed or subsampled images.");
       return(0);
    }
    if ((strncmp(Object,"^CLOUD_IMAGE",12)  == 0) ||
        (strncmp(Object,"^MOSAIC_IMAGE",13) == 0))
    {
       maxlat =  55.0;
       maxlon = 140.0;
       delta  =   5.0;
       start  =   1;
       incr   =  71;
       defdn  = numDN-1;
    }
    else
    {
       delta = 1.0;
       start =   7;
       incr  = 100;
       defdn =   0;
    }
    if (maxlat == 0.0 && minlat == 0.0 && maxlon == 0.0 && minlon == 0.0)
    {
        GetKeywordReal (CommandString, "MAXLAT", 0.0, &maxlat, &flag);
        GetKeywordReal (CommandString, "MAXLON", 0.0, &maxlon, &flag);
    }
    if (maxlon < 0.0)
        maxlon *= -1.0; /* convert to west lon */
    if (minlon < 0.0)
        minlon *= -1.0;
    if (minlon > maxlon)
    {
        hold   = maxlon;
        maxlon = minlon;
        minlon = hold;
    }

    GetKeywordString (CommandString, "LEG", "",label,&flag);
    if (flag >= 0)
    {
       j = 120;
       hold = -1.0;
       TextLine = j+8; TextSample = 540;
       sprintf( label, "%4.1f", hold);
       WriteText(label);
       hold += .5;
       DrawLine( j, 590-5, j++, 610, 35/(256/numDN));

       for (i=38;i<255;i+=3)
       {
           if (i==74 || i==116 || i == 158 || i == 200 || i == 242)
           {
               offset = 5;
               TextLine = j+8; TextSample = 540;
               sprintf(label,"%4.1f",hold);
               WriteText(label);
               hold += .5;
           }
           DrawLine( j, 590-offset, j++, 610, i/(256/numDN));
           DrawLine( j, 590,        j++, 610, i/(256/numDN));
           offset = 0;
       }
       DrawLine( j, 590, j++, 610, 255/(256/numDN));
       TextLine = j+20; TextSample = 520;
       strcpy(label,"Pigment");
       WriteText(label);
       TextLine = j+40; TextSample = 520;
       strcpy(label,"Concentratn");
       WriteText(label);
       TextLine = j+60; TextSample = 520;
       strcpy(label,"LOG(mg/m-3)");
       WriteText(label);
    }
    GetKeywordInteger (CommandString, "DN",     0, &dn,     &flag);
    if (flag <=0)
       dn = defdn;
    mlat = maxlat; mlon = maxlon;
    x = min(ns,dispns); y = min(nl-sl+1,dispnl);
    if (sl > 1)
    {
       if (sl < start)
           offset = 1-sl;
       else
       {
           i       = sl/incr+1;
           offset += i * incr - sl + 1;
           mlat   -= (float)i * delta;
       }
    }
    for (i=start;i<ns;i+=incr)     /* draw x axis */
    {
       DrawLine(1,i,y,i,dn);
       TextLine = start+offset+38; TextSample = i+1;
       sprintf(label,"%4.0f",mlon);
       WriteText(label);
       mlon -= delta;
    }

    for (i=start+offset;i<y;i+=incr)     /* draw y axis */
    {
       DrawLine(i,1,i,x,dn);
       TextLine = i+15; TextSample = 8;
       sprintf(label,"%4.0f",mlat);
       WriteText(label);
       mlat -= delta;
    } 
}


int DoSerial(void)
{
     TextLine = dispnl;  TextSample = 1;
     WriteText ("Not implemented yet.");
}

int DoPrint(void)
{
     TextLine = dispnl;  TextSample = 1;
     WriteText ("Not implemented yet.");
}

int DoLog(void)
{
    int    flag;
    char   LogName[40];

    GetKeywordSubcommand (CommandString, "CLO", &flag);
    if (flag >= 0)
    {
       fflush( LogFile );
       fclose( LogFile );
       logflag = 0;
       return(0);
    }
    GetKeywordSubcommand (CommandString, "OFF", &flag);
    if (flag >= 0)
    {
       fflush( LogFile );
       fclose( LogFile );
       logflag = 0;
       return(0);
    }
    GetKeywordSubcommand (CommandString, "ON", &flag);
    if (flag >= 0)
    {
       strcpy( LogName, "IMDISP.LOG" );
       LogFile = fopen( LogName, "wt" );
       if (LogFile != NULL)
          logflag = 1;
       else
          StatusLine(0, "Couldn't open log file");
       return(0);
    }
    GetKeywordString (CommandString, "OPE", "IMDISP.LOG", LogName, &flag);
    if (flag >= 0)
    {
       LogFile = fopen( LogName, "wt" );
       if (LogFile != NULL)
          logflag = 1;
       else
          StatusLine(0, "Couldn't open log file");
    }
    return(0);
}


