/***  IMDISP module MEM.C

      Extended/Expanded Memory Routines

    Contains high level memory routines to allocate, read and write, and free
    extended and expanded memory

    Separate module - 7/91 by Ron Baalke
***/

#define __MSC

/* * * * INCLUDE files * * * */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include "imdef.h"
#include "xmsif.h"
#include "emsif.h"

/* * * * External functions * * * */

int Memory_Alloc(unsigned long, int *, int *);
int Memory_PutLine(int, int, unsigned char *, long, int);
int Memory_GetLine(int, int, unsigned char *, long, int);
int Memory_Free(int,int);

/* * * * Global Variables * * * */

int XMS_Initialized = 0;
int EMS_Initialized = 0;

/*************************************************************************/
/* Memory_Alloc                                                          */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* This routine will allocate either extended or expanded memory.  It    */
/* will first check attempt to allocate extended memory, and if          */
/* unsuccessful, then well attempt expanded memory.  The handle and      */
/* which type of memory was allocated is returned.                       */
/*************************************************************************/

int Memory_Alloc(unsigned long size, int *handle, int *type)
{
   int error = 0;
   *type = NOWHERE;

   /* Attempt extended memory first */

   if (!XMS_Initialized && XMMlibinit() != 0)
      error = 1;
   if (!error)
   {
      XMS_Initialized = 1;
      *handle = XMMalloc(size);
      if(*handle == XMMOOPS)
         error = 1;
      else
         *type = EXTENDED_MEMORY;
   }

   /* Attempt expanded memory if extended memory attempt was no good */

   if (*type == NOWHERE)
   {
      error = 0;
      if (!EMS_Initialized && EMMlibinit() != 0)
         error = 1;
      if (!error)
      {
         EMS_Initialized = 1;
         *handle = EMMalloc(size);
         if(*handle == EMMOOPS)
            error = 1;
         else
            *type = EXPANDED_MEMORY;
      }
   }

   return(error);
}

/*************************************************************************/
/* Memory_PutLine                                                        */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* This routine will move a buffer to extended/expanded memory.  The     */
/* memory must have already been allocated by a call to Memory_Alloc.    */
/*                                                                       */
/*************************************************************************/

Memory_PutLine(handle,type,p,offset,length)
int  handle;
int  type;
unsigned char *p;
long offset;
int  length;
{
   int error = 0;

   if (type == EXTENDED_MEMORY)
      if (XMMcopyto((unsigned long) length, (unsigned char far *) p,
                    handle, offset) == XMMOOPS)
         error = 1;
   else if (type == EXPANDED_MEMORY)
      if (EMMcopyto((unsigned long) length, (unsigned char far *) p,
                    handle, offset) == EMMOOPS)
         error = 1;

   return(error);
}

/*************************************************************************/
/* Memory_GetLine                                                        */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* This routine will retrieve a buffer from extended/expanded memory.    */
/* The memory must have already been allocated by a call to Memory_Alloc.*/
/*                                                                       */
/*************************************************************************/

Memory_GetLine(handle,type,p,offset,length)
int handle;
int type;
unsigned char *p;
long offset;
int  length;
{
   int error = 0;
   if (type == EXTENDED_MEMORY)
      if (XMMcopyfrom((unsigned long) length, handle, offset,
                      (unsigned char far *) p) == XMMOOPS)
      error = 1;
   else if (type == EXPANDED_MEMORY)
      if (EMMcopyfrom((unsigned long) length, handle, offset,
                      (unsigned char far *) p) == EMMOOPS)
      error = 1;

   return(error);
}

/*************************************************************************/
/* Memory_Free                                                           */
/*                                                                       */
/* Written by Ron Baalke - 07/91                                         */
/*                                                                       */
/* This routine will free extended/expanded memory.                      */
/*                                                                       */
/*************************************************************************/

Memory_Free(int handle, int type)
{
   if      (type == EXTENDED_MEMORY)
      XMMfree(handle);
   else if (type == EXPANDED_MEMORY)
      EMMfree(handle);
}
